\\ Copyright 2015, 2016, 2017, 2018, 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ http://user42.tuxfamily.org/pari-geom/index.html

\\ Usage: read("geom.gp");
\\        geom_area_of_triangle(3+4*I, 1+7*I)  \\ == 17/2
\\
\\
\\ This is some functions for simple mostly 2-dimensional geometry such as
\\ areas, centroids, boundaries of triangles, polygons, etc.  The further
\\ geom-inertia.gp has things for mass moment of inertia.  addhelp() strings
\\ have docs of each function and see general notes in the comments below.
\\
\\     x = geom_point_x(p)                       coordinates
\\     y = geom_point_y(p)
\\     z = geom_point_z(p)
\\
\\     d = geom_abs(p)                           distance to origin
\\     norm = geom_norm(p)                       distance squared
\\     z = geom_sqrt_as_quad(n)
\\
\\     q = geom_quadrant_of_point(p)
\\     angle = geom_arg_positive(p)
\\     d = geom_dot_product(p1,p2)
\\
\\     len = geom_length_of_segment(p1,p2)       between points
\\     len = geom_length_of_path(v)
\\     len = geom_length_of_polygon(v)
\\
\\     s = geom_xydir_cmp(p1,p2)
\\     half_p = geom_xydir_half(p)               half angle
\\     p_rot = geom_point_rotate_xydir(p,xydir)
\\     v_rot = geom_points_rotate_xydir(v,xydir)
\\
\\     area = geom_area_of_triangle(p1,p2,p3=0)
\\     area = geom_area_of_rectangle5(b1,b2,r,t,l)
\\     area = geom_area_of_polygon(v)
\\     area = geom_area_of_intformal_rows(ymin,ymax, xmin,xmax)
\\     area = geom_area_of_circle(r)
\\     area = geom_area_of_ellipse(a,b)
\\
\\     p = geom_centroid_of_points(v)            unweighted mean
\\     p = geom_centroid_of_triangle(p1,p2,p3=0)
\\     p = geom_centroid_of_polygon(v)
\\
\\     v = geom_sides_of_polygon(v)
\\     s = geom_minrect_side_of_polygon(v)
\\     area = geom_minrect_area_of_polygon(v)
\\     area = geom_minrect_area_of_points_at_xydir(v,p)
\\
\\
\\ Points
\\ ------
\\
\\ Points p in the XY plane are taken as any of
\\
\\     complex numbers   3+4*I
\\     complex quads     quadgen(-12)
\\     [x,y]             Vec or Col
\\     [x;y]             Mat 2-element column
\\
\\ A few functions only accept complex numbers (and complex quads).  Most
\\ functions taking multiple points will need them as all complex numbers or
\\ all vectors (not a mixture).
\\
\\ When quads are given, the calculations are kept in quads so results are
\\ exact.  As of Pari/GP circa 2.11, quads cannot be intermixed with complex
\\ numbers or quads of different discriminant, so points will usually need
\\ to be compatible types.
\\
\\ Points as 2 element vectors can have both X and Y as real quads.  This
\\ sometimes helps for coordinates of otherwise incompatible discriminant,
\\ though some functions will still end up mixing them.
\\
\\ Functions with simple formulas allow polynomials for points, eg.
\\ 
\\     \\ symbols 'b and 'h as triangle base and height
\\     geom_area_of_triangle(['b,0],[0,'h]) == 1/2*'b*'h
\\
\\ This can get a formula for some polygon shape with a couple of parameters
\\ for extents.  Another possibility is to use a variable as placeholder for
\\ a number (a real) which cannot be exactly represented, eg. a cube root.
\\
\\ There's partial support for 3-element vectors for 3-D coordinates.  Area
\\ calculations only work in 2-D XY plane.  Segment lengths and some inertia
\\ works in 3-D.
\\
\\     z = geom_point_z(p)
\\     [x,y,z]           Vec or Col
\\     [x;y;z]           Mat 3-element column
\\
\\
\\ Speed
\\ -----
\\
\\ geom_point_x() etc coordinates inspect the type of each point (complex,
\\ quad, vector) each time so won't be particularly fast on large numbers of
\\ points, but they have the attraction of being flexible.
\\
\\ geom_point_y() of a complex quad returns a real quad for the imaginary
\\ part.  For very large discriminants this can be slow due to factorizing
\\ to find the discriminant square-free part.  If this occurs then it may be
\\ better to points as pairs [x,y], effectively pre-calculating the y.
\\
\\
\\ Requirements
\\ ------------
\\
\\ Designed for GP 2.7 up.  Clean to option "strictargs".
\\ No need for dynamic library install().
\\
\\
\\ Changes
\\ -------
\\
\\ Version 1 - the first version.
\\ Version 2 - misc more, including quadrant.
\\ Version 3 - ...


\\-----------------------------------------------------------------------------
\\ Miscellaneous

geom_sqrt_as_quad(n) =
{
  if(n=simplify(n),  \\ entirely real complex or quad
     iferr(my(den = denominator(n));
           my(v = core(numerator(n)*den,1));  \\ v = [squarefree, squares]
           v[2] * if(v[1]==1,1, quadgen(4*v[1])) / den,
           e,
           sqrt(n)),  \\ floats, complex
     0); \\ n==0
}
{
addhelp(geom_sqrt_as_quad,
"z = geom_sqrt_as_quad(n)
Return sqrt(n) in the form of a quad.
n can be an integer or rational, either positive, negative or zero.
For example geom_sqrt_as_quad(75) == 5*quadgen(12), since quadgen(12) is an
exact sqrt(3).

The quad is formed by finding the square-free kernel of n using core().
This might be slow on large n.

An automatic simplify() allows complex or quad n with 0 imaginary part.
But complex or quad perfect squares are currently not noticed.");
}

geom_INTERNAL_hypot(x,y,z=0) =
{
  \\ with simplify to allow squares of different x,y,x=sqrt to be mixed
  geom_sqrt_as_quad(simplify(sqr(x)) + simplify(sqr(y)) + simplify(sqr(z)));
}


\\-----------------------------------------------------------------------------
\\ Coordinates

\\ geom_INTERNAL_imag_of_quad(z) and geom_INTERNAL_imag_of_quad(z) take quad
\\ z and return the real or imaginary parts as further quads.
\\
\\ The real part includes the real part arising when z.disc == 1 mod 4.  The
\\ imaginary part is a new real quad, so for example sqrt3i=quadgen(-12)
\\ gives imaginary part sqrt3=quadgen(12).  When z.disc==1mod4 the new
\\ discriminant is found by geom_sqrt_as_quad() so might be slow on very
\\ large disc.
\\
\\     a + sqrt(disc)       a=0 if disc==0 mod 4
\\ z = --------------       a=1 if disc==1 mod 4
\\          2
\\ disc==0 mod 4, go to -disc
\\ disc==1 mod 4 have -disc == 3 mod 4, find its square-free part
\\
\\ vector(40,d,d--;d*=4;d+=1;d=-d; d/core(-d))
\\ vector(40,d,d--;d*=4;d=-d;d+=-3; (-d)%3)
\\ quadgen(-7)
\\ geom_sqrt_as_quad(7).disc == 28      \\ 4*7 == 28
\\
geom_INTERNAL_imag_of_quad(z) =
{
  my(d = z.disc);
  if(d>=0,return(0));         \\ disc>=0 is z real
  if(d%4,
     imag(z)/2 * geom_sqrt_as_quad(-d),
     imag(z) * quadgen(-d));
}
geom_INTERNAL_real_of_quad(z) =
{
  my(d = z.disc);
  if(d>=0, z, real(z) + if(d%4,imag(z)/2));
}

geom_point_x(p) =
{
  p = simplify(p);
  my(t = type(p));
  if(t=="t_QUAD", geom_INTERNAL_real_of_quad(p),

     t=="t_VEC" || t=="t_COL",
     if(#p==2 || #p==3, p[1],
        error("point as vector should be 2 or 3 elements")),

     t=="t_MAT",
     my(s=matsize(p));
     if(s==[2,1] || s==[3,1], p[1,1],
        error("point as matrix should be 2 or 3 element column")),

     t=="t_POL",
     Pol(apply(geom_point_x,Vec(p)),variable(p)),

     t=="t_RFRAC",
     my(r=real(p),i=imag(p));
     r + if(i, geom_point_x((p-r)/i)*i),

     real(p));
}
{
addhelp(geom_point_x,
"x = geom_point_x(p)
Return the X coordinate of a point p.  p can be
 real:    x=p
 complex: x=real(p)
 quad:    x=p if disc>=0 (entirely real) or add real part of a/2 if disc<0.
 vector or column: x=p[1]
 matrix column:    x=p[1,1]");
}

geom_point_y(p) = {
  p = simplify(p);
  my(t = type(p));
  if(t=="t_QUAD", geom_INTERNAL_imag_of_quad(p),

     t=="t_VEC" || t=="t_COL",
     if(#p==2 || #p==3, p[2],
        error("point as vector should be 2 or 3 elements")),

     t=="t_MAT",
     my(s=matsize(p));
     if(s==[2,1] || s==[3,1], p[2,1],
        error("point as matrix should be 2 or 3 element column")),

     t=="t_POL",
     Pol(apply(geom_point_y,Vec(p)),variable(p)),

     t=="t_RFRAC",
     my(i=imag(p));
     if(i, i*geom_point_y((p-real(p))/i)),

     imag(p));
}
{
addhelp(geom_point_y,
"y = geom_point_y(p)
Return the Y coordinate of a point p.  p can be
 real:    y=0
 complex: y=imag(p)
 quad:    y=0 if disc>=0 (real) or new quad for imaginary part if disc<0.
 vector or column: y=p[2]
 matrix column:    y=p[2,1]

For imaginary quads a new real quad is formed to exactly represent the
imaginary part.  For example sqrt3i = quadgen(-12) is sqrt(3)*I and
geom_point_y(sqrt3i) == quadgen(12) which is an exact sqrt(3).

For imaginary quads discriminant D==1 mod 4 the square-free kernel of D is
found with core() as per geom_sqrt_as_quad() and so might be a slow for
large D.");
}

geom_point_z(p) =
{
  p = simplify(p);
  my(t = type(p));
  if(t=="t_VEC" || t=="t_COL",
     if(#p==2, 0,
        #p==3, p[3],
        error("point as vector should be 2 or 3 elements")),

     t=="t_MAT",
     my(s=matsize(p));
     if(s==[2,1], 0,
        s==[3,1], p[3,1],
        error("point as matrix should be 2 or 3 element column")),

     0);
}
{
addhelp(geom_point_z,
"z = geom_point_z(p)
Return the Z coordinate of a point p.  p can be
 real, complex, quad: z=0
 vector or column:    z=p[3]
 matrix column:       z=p[3,1]");
}

geom_norm(p) =
{
  \\ with simplify to allow squares of different x,y,x=sqrt to be mixed
  simplify(sqr(geom_point_x(p)))
  + simplify(sqr(geom_point_y(p)))
  + simplify(sqr(geom_point_z(p)));
}
{
addhelp(geom_norm,
"norm = geom_norm(p)
Return the norm of point p, being x^2+y^2+z^2.");
}

geom_abs(p) =
{
  \\ If only one of x,y,z is non-zero then it is returned without going
  \\ through the square and sqrt.  Is that a good idea?
  my(x = geom_point_x(p),
     y = geom_point_y(p),
     z = geom_point_z(p));
  if((x!=0)+(y!=0)+(z!=0) <= 1,
     if(x!=0,sign(x)*x,
        y!=0,sign(y)*y,
        sign(z)*z),
     geom_INTERNAL_hypot(x,y,z));
}
{
addhelp(geom_abs,
"d = geom_abs(p)
Return the absolute value of point p, being the distance from the origin,
sqrt(geom_norm(p)) = sqrt(x^2+y^2+z^2).");
}


\\-----------------------------------------------------------------------------
\\ Quadrants

geom_quadrant_of_point(p) =
{
  my(x=geom_point_x(p),
     y=geom_point_y(p));
  if(x>0  && y>=0, 1,
     x<=0 && y>0,  2,
     x<0  && y<=0, 3,
     x>=0 && y<0,  4,
     0);
}
{
addhelp(geom_quadrant_of_point,
"q = geom_quadrant_of_point(p)
Return the quadrant number containing point p.
Quadrants are numbered 1 for x>0,y>0, then anti-clockwise around 2 for
x<0,y>0, etc.  Only the x,y parts of p are used, any z is ignored.

Quadrant boundaries are treated by x>0 axis in quadrant 1, then y>0 axis in
quadrant 2, etc, so each quadrant includes its starting axis, and next axis
in next quadrant (anti-clockwise around).  If p is the origin x=0,y=0 then
return quadrant 0.");
}

geom_arg_positive(p) =
{
  my(a=arg(geom_point_x(p)*1.0 + (1.0*I)*geom_point_y(p)));
  if(a>=0, a, a + 2*Pi);
}
{
addhelp(geom_arg_positive,
"angle = geom_arg_positive(p)
Return the angle in radians of point p in the X,Y plane, measured
anti-clockwise around from the X axis and in the range 0 <= angle < 2*Pi.

This is like arg() but returning a positive angle, and accepting the geom
style vector points.");
}


\\-----------------------------------------------------------------------------
\\ XY Directions

geom_xydir_half(p) = 
{
  my(y = geom_point_y(p),
     x = geom_point_x(p),
     h = geom_INTERNAL_hypot(x,y));
  if(x>=0, [x+h,y],
     y==0, [0,-x],     \\ x<0,y==0 go to +90deg
     [y,h-x]*sign(y));
}
{
addhelp(geom_xydir_half,
"half_p = geom_xydir_half(p)
p is a point in the XY plane.
Return a point in direction which is half-angle towards the X axis
(bi-section of the angle).

For example [1,1] is 45 degrees and half is 22.5 degrees,
  geom_xydir_half([1,1]) == [1+geom_sqrt_as_quad(2), 1]

Exactly what point is returned in the new direction is unspecified.

The intention is to stay exact as long as possible.  The current
implementation is Euclid's [x+h,y] where h=sqrt(x^2+y^2), and the usual
complex sqrt approach [y,h-x]*sign(y) when x<0.  The latter is a multiply by
(h-x)/y and avoids bad cancellation of floating point x+h when y is small so
that h is close to -x.");
}

geom_xydir_cmp(p1,p2) =
{
  my(c=lex(geom_quadrant_of_point(p1), geom_quadrant_of_point(p2)));
  if(c, c, lex(0,geom_area_of_triangle(p1,p2)));
}
{
addhelp(geom_xydir_cmp,
"s = geom_xydir_cmp(p1,p2)
p1 and p2 are points in the x,y plane representing directions.
Return sign s = -1,0,1 according as comparison p1 cmp p2 of their angles
measured anti-clockwise around from the X axis in the range 0 <= a < 2*Pi.

s is like lex(), the sign of an imagined angle(p1) - angle(p2) subtraction.
Only directions of p1,p2 are used, their magnitudes are ignored.

p1 or p2 at the origin 0 (so not having an angle at all) is reckoned smaller
than a non-origin point.");
}


\\-----------------------------------------------------------------------------
\\ Points

geom_dot_product(p1,p2) =
{
    simplify(geom_point_x(p1) * geom_point_x(p2))
  + simplify(geom_point_y(p1) * geom_point_y(p2))
  + simplify(geom_point_z(p1) * geom_point_z(p2));
}
{
addhelp(geom_dot_product,
"d = geom_dot_product(p1,p2)
Return the dot product of points p1 and p2.
This is their coordinates x1*x2 + y1*y2 + z1*z2.");
}

\\ (x+y*I)*(xd+yd*I) == (xd*x-yd*y) + (xd*y+yd*x)*I
geom_point_rotate_xydir(p,xydir) = {
  my(x = geom_point_x(p),
     y = geom_point_y(p),
     xd = geom_point_x(xydir),
     yd = geom_point_y(xydir));
  [x*xd - y*yd, x*yd + y*xd] / geom_abs(xydir);
}
{
addhelp(geom_point_rotate_xydir,
"p_rot = geom_point_rotate_xydir(p,xydir)
p and xydir are points in the XY plane.
Return p rotated by the direction xydir.
The magnitude of p (distance from the origin) is unchanged.
xydir is used only for direction, its magnitude is ignored.");
}

geom_points_rotate_xydir(v,xydir) = \
  apply(p->geom_point_rotate_xydir(p,xydir), v);
{
addhelp(geom_points_rotate_xydir,
"v_rot = geom_points_rotate_xydir(v,xydir)
v is a vector of points and xydir is a point in the XY plane.
Return v rotated by the direction xydir in the manner of
geom_point_rotate_xydir().");
}

geom_centroid_of_points(v) = vecsum(v) / #v;
{
addhelp(geom_centroid_of_points,
"p = geom_centroid_of_points(v)
v is a vector of points which are taken as having the same mass each.
Return the centroid (centre of gravity) of the points.
This is simply the mean of all the points in v.
Must have at least 1 point in v.");
}

geom_minrect_area_of_points_at_xydir(v,p) =
{
  v /= p;
  my(vx = apply(geom_point_x,v),
     vy = apply(geom_point_y,v));
  norm(p) * (vecmax(vx) - vecmin(vx))
          * (vecmax(vy) - vecmin(vy));
}
{
addhelp(geom_minrect_area_of_points_at_xydir,
"area = geom_minrect_area_of_points_at_xydir(v,p)
v is a vector of complex number points in the XY plane.
p is a complex number point in the XY plane giving a direction.
Must have p!=0.

Return the area of the minimum area rectangle with one side oriented in the
direction of p enclosing all points of v.  This means finding the widest
points in the direction of p and direction perpendicular to p.");
}


\\-----------------------------------------------------------------------------
\\ Line Segments

geom_length_of_segment(p1,p2) =
{
  geom_INTERNAL_hypot(geom_point_x(p1) - geom_point_x(p2),
                      geom_point_y(p1) - geom_point_y(p2),
                      geom_point_z(p1) - geom_point_z(p2));
}
{
addhelp(geom_length_of_segment,
"len = geom_length_of_segment(p1,p2)
Return the length of the line segment from points p1 to p2.
If p1 and p2 are exact then the result is represented exactly in a quad (of
some suitable discriminant).");
}

geom_length_of_path(v) = \
  sum(i=1,#v-1, geom_length_of_segment(v[i], v[i+1]));
{
addhelp(geom_length_of_path,
"len = geom_length_of_path(v)
v is a vector of points representing a path.
Return the total length of the path, being simply sum of
geom_length_of_segment() between points.");
}


\\-----------------------------------------------------------------------------
\\ Rectangles

geom_area_of_rectangle5(b1,b2,r,t,l) =
{
  my(d=b2-b1);
  geom_point_y((t-b1)/d) * geom_point_x((r-l)/d) * norm(d);
}
{
addhelp(geom_area_of_rectangle5,
"area = geom_area_of_rectangle5(b1,b2,r,t,l)
b1,b2,r,t,l are five complex number points in the XY plane defining a
rectangle.  Return the area of the rectangle.

    *-----t-----*
    |           |
    l           r
    |           |
    *--b1---b2--*

b1,b2 define a baseline.  r gives the right side by a point on a line at 90
degrees to the baseline.  t gives the top side by a point on a line parallel
to the baseline.  l gives the left side by a point on a line at 90 degrees
to the baseline.

Must have b1 != b2 so that they define a baseline.  They can be given either
way around b1,b2 or b2,b1.  They can be out wider than l,r too, only the
projection of r and l on the baseline is used.  Points r,t,l determine the
other sides.  They must go anti-clockwise (+90 degrees each time) around.
The actual orientation in the plane is arbitrary as long as the sides go
anti-clockwise around.  If you have r,l the wrong way around then the area
will be negative.

Usually area of a rectangle is simply width*height.  This function can be
used when a rectangle is known by points on its sides, not its corners.");
}


\\-----------------------------------------------------------------------------
\\ Triangles

geom_INTERNAL_point_sub(p1,p2) =
{
  iferr(p1-p2,
        e,
        [geom_point_x(p1) - geom_point_x(p2),
         geom_point_y(p1) - geom_point_y(p2),
         geom_point_z(p1) - geom_point_z(p2)]);
}

\\ imag(conj('xp1+I*'yp1)*('xp2+I*'yp2)) == xp1*yp2 - yp1*xp2
geom_area_of_triangle(p1,p2,p3=0) = 
{
  p1=geom_INTERNAL_point_sub(p1,p3);
  p2=geom_INTERNAL_point_sub(p2,p3);
  (  geom_point_x(p1)*geom_point_y(p2)
   - geom_point_y(p1)*geom_point_x(p2) ) / 2;

  \\ p1 -= p3;
  \\ p2 -= p3;
  \\ geom_abs(geom_INTERNAL_cross_product(p1,p2)) / 2;
  \\ 1/2 * geom_point_y(conj(p1)*p2);     \\ complex numbers
}
{
addhelp(geom_area_of_triangle,
"area = geom_area_of_triangle(p1,p2,p3=0)
p1,p2,p3 are points which are vertices of a triangle in the XY plane.
Return the area of the triangle.

p1,p2,p3 should be anti-clockwise around (the usual mathematical direction).
If they're clockwise then the area returned is negative.  Negative area
like that is good when summing triangles of a polygon or similar for net
area inside.

p3 defaults to 0 which is a triangle given by 2 points (anti-clockwise) and
the origin.");
}

geom_centroid_of_triangle(p1,p2,p3=0) = (p1 + p2 + p3)/3;
{
addhelp(geom_centroid_of_triangle,
"p = geom_centroid_of_triangle(p1,p2,p3=0)
Return the centroid (centre of gravity) point in the triangle given by
vertex points p1,p2,p3.

The centroid of a triangle is simply the mean of the vertices (p1+p2+p3)/3.
p3 defaults to 0 which is a triangle given by 2 points and the origin.");
}


\\-----------------------------------------------------------------------------
\\ Polygons

\\ cf Charles R Greathouse charles.gp polygonArea() taking x,y successively
\\ in a vector and area by triangles.
\\
geom_area_of_polygon(v) =
{
  if(#v<3, 0,
     sum(i=1,#v-1, geom_area_of_triangle(v[i],v[i+1]))
     + geom_area_of_triangle(v[#v],v[1]));
}
{
addhelp(geom_area_of_polygon,
"area = geom_area_of_polygon(v)
v is a polygon in the form of a vector of points.
Return the area of the polygon.

The points in v should go anti-clockwise around (the usual mathematical
direction).  If they're clockwise then the area returned is negative (which
can be useful for getting a net area).");
}

geom_centroid_of_polygon(v) =
{
  my(weighted_point=0, total_area=0);
  for(i=1,#v,
     my(p1 = v[i],
        p2 = v[(i%#v)+1],
        area = geom_area_of_triangle(p1,p2));
     total_area += area;
     weighted_point += area*geom_centroid_of_triangle(p1,p2);
     \\ print(p1" "p2"  area="area" tcent="geom_centroid_of_triangle(p1,p2));
  );
  weighted_point / total_area;
}
{
addhelp(geom_centroid_of_polygon,
"p = geom_centroid_of_polygon(v)
v is a polygon in the form of a vector of points going anti-clockwise.
Return the centroid (centre of gravity) of the polygon.
The area of the polygon must be non-zero.");
}

\\ geom_INTERNAL_minrect_all_of_polygon(v) returns vector
\\ [area,side,side,...] which is the minimum rectangle area and all the side
\\ numbers to which such a minimum area rectangle are aligned.
geom_INTERNAL_minrect_all_of_polygon(v) =
{
  if(#v<1,[0,1]);
  my(l=List([]));    \\ [area, side,side,...]
  for(s=1,#v,
      my(p1 = v[s],
         p2 = v[if(s==#v,1,s+1)],
         delta=p1-p2);
     delta || next;  \\ allow for empty sides in v
     my(area = geom_minrect_area_of_points_at_xydir(v, delta));
      if(!#l || area < l[1],  l=List([area,s]),   \\ new minimum
         area == l[1],        listput(l,s)));     \\ an equal minimum
  Vec(l);
}

geom_minrect_side_of_polygon(v) = \
  geom_INTERNAL_minrect_all_of_polygon(v)[2];
{
addhelp(geom_minrect_side_of_polygon,
"s = geom_minrect_side_of_polygon(v)
v is a convex polygon in the form of a vector of points going anti-clockwise.
Return the first side which is aligned to a minimum area rectangle around
the polygon.  For example return 3 means side v[3] to v[4].

A minimum area rectangle is always aligned to at least one side of the
polygon.  There might be other sides aligned to this rectangle, and there
might be other equal minimum area rectangles at different orientations.");
}

geom_minrect_area_of_polygon(v) = \
  geom_INTERNAL_minrect_all_of_polygon(v)[1];
{
addhelp(geom_minrect_area_of_polygon,
"area = geom_minrect_area_of_polygon(v)
v is a convex polygon in the form of a vector of points going anti-clockwise.
Return the area of the minimum area rectangle enclosing the polygon.");
}

geom_sides_of_polygon(v) = \
  vector(#v,i, geom_INTERNAL_point_sub(v[if(i==#v,1,i+1)], v[i]));
{
addhelp(geom_sides_of_polygon,
"v = geom_sides_of_polygon(v)
v is a polygon in the form of a vector of complex number points.
Return the sides of the polygon as differences
  [ v[2]-v[1], v[3]-v[2], ..., v[#v]-v[#v-1], v[1]-v[#v] ].");
}

geom_length_of_polygon(v) =
{
  v=apply(geom_abs,geom_sides_of_polygon(v));
  iferr(vecsum(v),e, vecsum(v*1.0));
}
{
addhelp(geom_length_of_polygon,
"len = geom_length_of_polygon(v)
v is a polygon in the form of a vector of complex number points.
Return the boundary length of the polygon.
This is exact if the sides all have compatible slopes (so that the
hypotenuses are represented with a single quad discriminant), or inexact
otherwise.");
}


\\-----------------------------------------------------------------------------
\\ intformals

geom_area_of_intformal_rows(ymin,ymax, xmin,xmax) =
{
  my(poly=intformal(xmax-xmin, 'y));
  subst(poly, 'y, ymax) - subst(poly, 'y, ymin);
}
{
addhelp(geom_area_of_intformal_rows,
"area = geom_area_of_intformal_rows(ymin,ymax, xmin,xmax)
Return the area of a shape in the XY plane defined by integrals over rows.

A given row goes from x=xmin to x=xmax.  xmin,xmax can be polynomials in the
Y coordinate variable 'y, so allowing shapes like parabolas etc.
Rows are integrated from y=ymin to y=ymax.

            y=ymax  
    area = integral  xmax - xmin  dy
            y=ymin

This uses intformal() so all of xmin,xmax, ymin,ymax can be polynomials in
other variables too, for example some 'w or 'h as an overall width or height
to parameterize the result.");
}


\\-----------------------------------------------------------------------------
\\ Circles and Ellipses

geom_area_of_circle(r) = Pi*sqr(r);
{
addhelp(geom_area_of_circle,
"area = geom_area_of_circle(r)
Return the area of a circle of radius r, being Pi*r^2.");
}

geom_area_of_ellipse(a,b) = Pi*a*b;
{
addhelp(geom_area_of_ellipse,
"area = geom_area_of_ellipse(a,b)
Return the area of an ellipse of semi-radii a,b, being Pi*a*b.");
}


\\-----------------------------------------------------------------------------

\\  LocalWords:  geom gp minrect intformal ymin ymax xmin xmax XY
\\  LocalWords:  Vec quadgen Pari imag iferr squarefree addhelp hypot abs
\\  LocalWords:  xydir rt matsize listput vecsum dir vecmax vecmin dy poly
\\  LocalWords:  subst parameterize RFRAC Ryde conj flonum bi
