#!/usr/bin/perl -w

# Copyright 2015, 2016, 2017, 2018, 2019, 2020, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Usage: perl my-check-tex.pl filename...
#
# Check for some mistakes in the given TeX or LaTeX files.

use 5.010;
use strict;
use warnings;
use File::Basename;
use File::Slurp;
use Getopt::Long;
use List::Util 'max','min','sum';
use Math::NumSeq::OEIS::File;
use Regexp::Common 'balanced';
use Tie::IxHash;

use FindBin;
use lib::abs $FindBin::Bin;

# uncomment this to run the ### lines
# use Smart::Comments;

our $VERSION = 0;
$|=1;

my $progname = $FindBin::Script;
my $verbose = 0;


# Data::Lock sets SvREADONLY but alas in perl circa 5.30 for some reason
# this is a 2x or more slowdown.  So may want it to check the string
# crunching isn't changing the original $tex_str etc, but only for
# development.
#
# use Data::Lock 'dlock';
sub dlock {}

# ENHANCE-ME: \% would be a literal percent, if odd num backslashes
my $comment_re = qr/%[^\n]*\n/;         # TeX comment % through to newline
my $comments_re = qr/($comment_re)*/o;  # zero or more TeX comments

# but just the start of a /* .. */ style
my $gp_comment_re = qr{\\\\|/\*};

# Regexp::Common::balanced
my $braces_re = $RE{'balanced'}{-parens=>'{}'};

{
  package Regexp::Common::my_experimental_tex;
  use Regexp::Common 'balanced';

  Regexp::Common::pattern
      (name   => ['TeX','integer_constant'],
       create => '[0-9]+');
  Regexp::Common::pattern
      (name   => ['TeX','octal_constant'],
       create => '[0-7]+');
  Regexp::Common::pattern
      (name   => ['TeX','hexadecimal_constant'],
       create => '[0-9A-F]+');

  # my $coerced_integer = "$internal_dimen|$internal_glue";
  # Regexp::Common::pattern
  #     (name   => ['TeX','coerced_integer'],
  #      create => $coerced_integer);

  my $normal_integer = '[0-9]+|\'[0-7]+|"[0-9A-F]+|`.';
  Regexp::Common::pattern
      (name   => ['TeX','normal_integer'],
       create => $normal_integer);

  # Regexp::Common::pattern
  #     (name   => ['TeX','normal_integer'],
  #      create => "$normal_integer|$coerced_integer");

  # em or ex
  my $internal_unit = 'e[mx]';
  Regexp::Common::pattern
      (name   => ['TeX','internal_unit'],
       create => $internal_unit);

  # pt, cm, etc
  my $physical_unit = 'bp|c[cm]|dd|p[ct]|in|m[mu]|sp';
  Regexp::Common::pattern
      (name   => ['TeX','physical_unit'],
       create => $physical_unit);

  # em, ex, pt, cm, etc
  my $unit = "$internal_unit|(?:true )?(?:$physical_unit)";
  Regexp::Common::pattern
      (name   => ['TeX','unit'],
       create => $unit);

  my $number = '[+-]?[ \t]*[0-9]*([0-9]|[.,])[0-9]*';
  Regexp::Common::pattern
      (name   => ['TeX','number'],
       create => $number);

  # eg: 1.5em
  my $dimen = "(?:$number)[ \t]*(?:$unit)";
  Regexp::Common::pattern
      (name   => ['TeX','dimen'],
       create => $dimen);

  my $kern_sans_backslash_re = "m?kern[ \t]*$dimen";
  Regexp::Common::pattern
      (name   => ['TeX','kern'],
       create => sub {
         my ($self, $flags) = @_;
         return (exists $flags->{'-nobackslash'} ? '' : '\\')
           . $kern_sans_backslash_re;
       });

  # Eg: [blah blah]
  # bracketed optional argument to a latex macro
  # balanced braces {blah} nest and parse as a unit,
  # otherwise the first ], so no nesting of [ ] (only inside braces)
  # -keep sets $1 to the part inside the brackets
  # -nobrackets matches without the brackets
  Regexp::Common::pattern
      (name   => ['LaTeX','optarg'],
       create => sub {
         my ($self, $flags) = @_;
         my $re = '(?:[^]{]|(?:' . $RE{'balanced'}{-parens=>'{}'} . '))*';
         if (exists $flags->{'-keep'}) { $re = "(?k:$re)"; }
         unless (exists $flags->{'-nobrackets'}) { $re = '\\[' . $re . '\\]'; }
         return $re;
       });

=pod

=for stopwords Ryde filename filenames basename Basename-only lockfile lockfiles symlink symlinks PID autosave builtin builtins non-builtin

=head1 NAME

Regexp::Common::TeX -- regexps for input TeX syntax

=for test_synopsis my ($str)

=head1 SYNOPSIS

 use Regexp::Common 'TeX', 'no_defaults';
 if ($str =~ /$RE{TeX}{dimen}/) {
    # 3em or .5pt etc ...
 }

 # regexp
 my $re1 = $RE{TeX}{number};

 # subroutine style to get regexp
 use Regexp::Common 'RE_TeX_physical_unit','no_defaults';
 my $re2 = RE_TeX_physical_unit();

=head1 DESCRIPTION

MUCH OF THIS IS INCOMPLETE OR SUBTLY WRONG

TeX also allows dimension register thingies like \dimen6 and specifies that
in the BNF syntax.  Not sure if would want that in regexps for completeness,
or would often find the physical measures plus em and ex enough.

This module is regexps matching input syntax used TeX, as defined by Knuth
in the TeX Book.  The names for various parts follow the BNF there.

See L<Regexp::Common> for basic operation of C<Regexp::Common>.

=head2 Patterns

=over

=item C<$RE{TeX}{internal_unit}>

Match a TeX "internal unit" which means simply two letters

    em
    ex

=item C<$RE{TeX}{physical_unit}>

Match a TeX "physical unit" which means simply two letter abbreviations

    bp
    cc
    cm
    dd
    pc
    pt
    in
    mm
    mu
    sp

=item C<$RE{TeX}{unit}>

Match a TeX unit, either C<internal_unit> or C<physical_unit>

=item C<$RE{TeX}{number}>

Match a TeX number.  This pattern is

    [+-]?
    [ \t]*
    [0-9]*
    ([0-9]|[.,])
    [0-9]*

So optional sign +/-, spaces and tabs permitted after it, then digits and
possible decimal point as either a dot "." or comma ",".  No digits at all
is permitted, as long as a . or , decimal.

=item C<$RE{TeX}{dimen}>

Match a TeX "dimen" which is

    number [ \t]* unit


=item C<$RE{LaTeX}{optarg}>

Match a LaTeX optional argument

    [something]

as for example \foo[this is optional].  In the "something" text, a further
open [ is permitted, but the argument ends at the first ], ie. no nesting of
brackets.

Braces groups {...} in the text do nest, and are matched that way here (per
L<Regexp::Common::balanced>).  Any brackets within a {...} group are
arbitrary.  So for instance an optional argument containing a macro call
also with an optional argument may have to be given

    \foo[{text \bar[option] text}]

A form like this is matched here, but don't be surprised if LaTeX demands a
few of levels of {{..}} braces before the text survives re-evaluation (and
even then not certain).

=back

=head1 IMPORTS

This module should be loaded through the C<Regexp::Common> mechanism, see
L<Regexp::Common/Loading specific sets of patterns.>.  Remember that loading
an add-on pattern like this module also loads all the builtin patterns by
default.

    # TeX plus all builtins
    use Regexp::Common 'TeX';

If you want only C<$RE{TeX}> then add C<no_defaults> (or list specific
desired builtins).

    # TeX alone
    use Regexp::Common 'TeX', 'no_defaults';

=head1 SEE ALSO

L<Regexp::Common>

=cut

}
# TeX Book chapter 24 page 270 dimens syntax spec
# my $internal_unit_re = qr/e[mx]/;
# my $physical_unit_re = qr/bp|c[cm]|dd|p[ct]|in|m[mu]|sp/;
# my $unit_re = qr/$internal_unit_re|(true )?($physical_unit_re)/o;
# my $number_re = qr/[+-]?[ \t]*[0-9]*([0-9]|[.,])[0-9]*/;
# my $dimen_re = qr/$number_re[ \t]*$unit_re/o;

my $internal_unit_re = qr/$RE{'TeX'}{'internal_unit'}/o;
my $physical_unit_re = qr/$RE{'TeX'}{'physical_unit'}/o;
my $unit_re          = qr/$RE{'TeX'}{'unit'}/o;
my $number_re        = qr/$RE{'TeX'}{'number'}/o;
my $dimen_re     = qr/$RE{'TeX'}{'dimen'}/o;
my $kern_sans_backslash_re = qr/m?kern[ \t]*(?:$dimen_re)/o;
my $kern_re = qr/\\$kern_sans_backslash_re/o;
{
  "em" =~ /^$unit_re$/ or die;
  "mu" =~ $unit_re or die;
  ".1" =~ /^$number_re$/ or die;
  "1.23" =~ $number_re or die;
  "1mu"  =~ $dimen_re or die;
  ".1em" =~ /^$dimen_re$/ or die;
  "mkern1mu" =~ $kern_sans_backslash_re or die;
  "\\mkern1mu" =~ $kern_re or die;

  ("[blah]" =~ $RE{LaTeX}{optarg} && $& eq '[blah]') or die;
  ("[blah]" =~ $RE{LaTeX}{optarg}{-keep} && $1 eq 'blah') or die;
  ("[bl]ah]" =~ $RE{LaTeX}{optarg}{-keep} && $1 eq 'bl') or die;
  ("foo]bar" =~ $RE{LaTeX}{optarg}{-keep}{-nobrackets} && $& eq 'foo') or die;
}

my $tex_filename;
my $bad = 0;
sub bad {
  my $linenum = shift;
  $bad = 1;
  print "$tex_filename:$linenum: ",@_,"\n";
}
sub bad_but_not_fatal {
  my $linenum = shift;
  print "$tex_filename:$linenum: ",@_,"\n";
}

sub tex_strip_comments {
  my ($str) = @_;
  # even number of backslashes  \\\\% ...
  # delete the % onwards, keep the backslashes
  $str =~ s/((^|[^\\])(\\\\)*)%[^\n]*/$1/mg;
  return $str;
}
CHECK {
  tex_strip_comments('% foo') eq '' or die;
  tex_strip_comments("% foo\nbar") eq "\nbar" or die;
  tex_strip_comments("a% foo\nbar") eq "a\nbar" or die;
  tex_strip_comments("a\\% foo") eq "a\\% foo" or die;
  tex_strip_comments("a\\\\% foo") eq "a\\\\" or die;
  tex_strip_comments("a\\\\\\% foo") eq "a\\\\\\% foo" or die;
}

sub verbose {
  my ($level, @output) = @_;
  if ($verbose >= 2) {
    require Devel::Mallinfo;
    my $h = Devel::Mallinfo::mallinfo();
    printf "   mallinfo arena %.1fmb + %.1fmb\n",
      $h->{'arena'} / 1e6, $h->{'hblkhd'} / 1e6;
  }
  if ($verbose >= $level) {
    print @output,"\n";
  }
}


sub check_file {
  dlock my $tex_str = File::Slurp::slurp($tex_filename);

  # comment is %, literal percent is \%
  # delete when preceded by even num backslashes
  my $tex_str_sans_comments = $tex_str;
  $tex_str_sans_comments =~ s/((^|[^\\])(\\\\)*)%[^\n]*/$1/mg;
  dlock $tex_str_sans_comments;

  # count_lines($tex_str) == count_lines($tex_str_sans_comments)
  #   or die "oops, line count changed";

  my $tex_str_sans_comments_and_tikz = $tex_str_sans_comments;
  my $count_tikz = 0;
  my $count_tikz_lines = 0;
  $tex_str_sans_comments_and_tikz
    =~ s{\\begin\{tikzpicture\}.*?\\end\{tikzpicture\}|\\tikzmath$braces_re}
        { my $str = $&;
          $count_tikz++;
          $count_tikz_lines += count_lines($str);
          str_leave_newlines($str)
        }geso;
  dlock $tex_str_sans_comments_and_tikz;

  verbose(1, "$tex_filename ",count_lines($tex_str)," lines,",
          " $count_tikz tikz of $count_tikz_lines lines");

  # all \hyphenation{foo-bar qu-ux}, with hyphens deleted for the keys
  verbose (2, "hyphenations");
  my %hyphenation;
  for ($tex_str_sans_comments) {
    while (/\\hyphenation\{(.*?)\}/g) {
      my $linenum = pos_linenum($_);
      my $h = $1;
      $h =~ s/^\s+//;
      $h =~ s/\s$//;
      foreach my $word (split /\s+/, $h) {
        $word =~ s/-//g;
        $hyphenation{$word} = $linenum;
      }
    }
  }

  #----------------------------------------------------------------------------
  # check for \vspace{.5\topsep} before align

  verbose(2, "environment vspace skip types");
  for my $str ($tex_str_sans_comments) {
    my %environ_to_skip = (align => 'abovedisplayskip',
                           center     => 'topsep',
                           flushright => 'topsep',
                          );
    while ($str =~ m<\\vspace\{[^}]\\(?<skip>[a-z]+)\}
                     \s*
                     \\begin\{(?<environ>[a-z]+)\}>xsg) {
      my $got_skip  = $+{'skip'};
      my $environ   = $+{'environ'};
      my $want_skip = $environ_to_skip{$+{'environ'}};
      if (! defined $want_skip) {
        my $linenum = pos_linenum($str);
        bad($linenum, "no skip type known for environment: $environ");
      } elsif ($got_skip ne $want_skip) {
        my $linenum = pos_linenum($str);
        bad($linenum, "environment $environ want $want_skip got $got_skip");
      }
    }
  }

  #----------------------------------------------------------------------------
  # check for multi-chars in maths mode

  verbose(2, "maths mode multi-char identifiers");
  {
    ### multi-char words in maths ...

    my $str = $tex_str;
    $str =~ s/^%[^\n]*/%/mg;  # whole line comments become %
    $str =~ s/([^\\\n](\\\\)*)%[^\n]*/$1/mg;  # non-whole-line comments delete
    # ### $str

    my %allow;
    while ($tex_str =~ /^% MY-CHECK-TEX-ALLOW-MULTICHAR: (.*)/mg) {
      ### $1
      @allow{split /\s+/, $1} = ();  # hash slice
    }

    my %maths_environments = (align       => 1,
                              aligned     => 1,
                              cases       => 1,
                              equation    => 1,
                              displaymath => 1,
                             );

    my @state = 0;  # 0=text, 1=maths, 2=tikz
    while ($str
           =~ m!
              \\(?<BeginEnd>begin|end)\{(?<Environment>[^}]*)\}
              |(?<Text>
                  \\(text(bf|sc)?
                    |math(it|rm|bf|sf|tt|cal)
                    |operatorname\*?
                    |raisebox\{[^}]*\}                  #\raisebox{1pt}{...
                    |index|kbd|var
                    |eqref|noeqref|label|intertext|tag\*?
                    |[hvm]space
                    |rule(\[[^]]*\])?\{[^}]*\}          #\rule[]{}{...
                    |genfrac\{[^}]*\}\{[^}]*\}\{[^}]*\} #\genfrac{}{}{}{...
                    |setlength\\[a-zA-Z]+               #\setlength\foo{...
                    )
                    \s*\{
                )
              |(?<Kern>
                  $kern_re                        # \kern.5em, \mkern1mu
                |\\\\\*?\[$RE{TeX}{dimen}\]       # \\[.5ex]
                |\\((thin|med|thick)muskip=|abovedisplayskip=)
                  $RE{TeX}{dimen}                 # \thickmuskip=1mu
                  )
              |(?<LineBreak>\\\\\*?(\[[a-z0-9.]+\])?)
              |(?<Backslash>\\)?         # possible macro marker
               [[:alpha:]][[:alnum:]]+   # text word, including X123 digits
              |(?<Underscore>[_^]
                  [^^_{}) \t\r\n\\]      # _\macro ok
                  [^^_{}()\[\] \t\r\n,\$=/.+-]+)
              |(?<Star>\*)
              |[\${}]
              |\n\n
              |\\[\$*]       # \$ not maths, \* not a star
              |\(\$          # tikz coordinate open
              |\$\)          # tikz coordinate close
               !igx) {
      my $match = $&;
      my $begin = $+{'BeginEnd'};
      my $environment = $+{'Environment'};
      my $underscore = $+{'Underscore'};
      my $star = $+{'Star'};
      my $kern = $+{'Kern'};
      my $text = $+{'Text'};
      my $backslash = $+{'Backslash'};
      my $linebreak = $+{'LineBreak'};

      ### $match
      if (defined $begin) {
        if ($begin eq 'begin') {
          ### begin environment group: $maths_environments{$environment}
          push @state, $maths_environments{$environment} ? 1 : 0;
        } else {
          ### end group ...
          pop @state;
        }
      } elsif (defined $text) {
        ### text mode open ...
        push @state, 0;

      } elsif (defined $kern) {
        ### kern etc: $kern
      } elsif (defined $backslash) {
        ### other backslash macro ...
      } elsif (defined $linebreak) {
        ### line break \\ macro ...

      } elsif ($match eq '$') {
        ### flip text/maths mode ...
        $state[-1] ^= 1;
      } elsif ($match eq "\n\n") {
        ### blank line, paragraph end ...
        if ($state[-1]) {
          ### oops ...
        }
        $state[-1] = 0;  # text mode at paragraph end


      } elsif ($match eq '{') {
        ### open group ...
        push @state, $state[-1];
      } elsif ($match eq '}') {
        ### close group ...
        pop @state;

      } elsif ($match eq '($') {
        ### tikz calc ...
        push @state, 2;

      } elsif (defined $underscore) {
        if ($state[-1] == 1) {
          my $linenum = pos_linenum($str);
          bad($linenum, "multi-char subscript or superscript: $match");
        }

      } elsif (defined $star) {
        if ($state[-1] == 1) {
          my $linenum = pos_linenum($str);
          bad($linenum, "star * in maths should be \\star usually");
        }

      } elsif ($state[-1]==1 && length($match) > 1) {
        unless (exists $allow{$match}) {
          my $linenum = pos_linenum($str);
          bad($linenum, "multi-char word in maths mode: $match");
        }
      }
    }
  }

  #----------------------------------------------------------------------------
  # check unused \bibitem{}

  verbose(2, "unused \\bibitem{}");
  for my $str ($tex_str_sans_comments) {

    tie my %bibitems, 'Tie::IxHash';  # in the order seen in the document
    while ($str =~ m<\\bibitem\{(?<bibitem>.*?)\}>g) {
      my $bibitem = $+{'bibitem'};
      ### $bibitem
      next if $bibitem =~ /#/;  # macro parameters
      my $linenum = pos_linenum($str);
      if (defined $bibitems{$bibitem}) {
        bad($linenum, "duplicate bibitem: $bibitem");
        bad($bibitems{$bibitem}, "was defined here");
      } else {
        $bibitems{$bibitem} = $linenum;
      }
    }

    # check all \cite are to a defined \bibitem
    my %used_bibitems;
    while ($str =~ m<\\cite($RE{LaTeX}{optarg})?\{(?<cite>.*?)\}>g) {
      my $cite_list = $+{'cite'};
      ### $cite_list
      next if $cite_list =~ /#/;  # macro parameters
      foreach my $cite (split /,/, $cite_list) {
        if (! exists $bibitems{$cite}) {
          my $linenum = pos_linenum($str);
          bad($linenum, "no such bibitem: $cite");
        }
        $used_bibitems{$cite} = 1;
      }
    }

    # warn on \bititem not used by some \cite
    foreach my $bibitem (keys %bibitems) {
      if (! $used_bibitems{$bibitem}) {
        bad_but_not_fatal($bibitems{$bibitem},
                          "unused bibitem: $bibitem");
      }
    }
  }

  #----------------------------------------------------------------------------
  # check for \MyAnumber[annotation]{A999999} where different annotation same A-number

  verbose (2, "MyAnumber annotations");
  {
    # believe latex goes up to the next "]", no nesting
    my %anum_to_annotation;  # 'A999999' => [$str, $linenum]
    my %annotation_to_anum;  # $str => ['A999999',$linenum]
    my @unannotated;
    for ($tex_str_sans_comments) {
      while (m{\\MyMulAnumber                        \s*
               (?<annotation>$RE{LaTeX}{optarg})?    \s*
               $braces_re                            \s*
               \{(?<anum>.*?)\}
             |
               (?<newcommand>\\newcommand)\\My(Fullsize|Index)?Anumber  \s*
               (?<annotation>$RE{LaTeX}{optarg})?    \s*
               \{(?<anum>.*?)\}
             }gxo) {
        next if defined $+{'newcommand'};
        my $annotation = $+{'annotation'} // '';
        my $anum = $+{'anum'};
        ### $anum
        $anum =~ s/^\{//;
        $anum =~ s/\}$//;
        $annotation =~ s/\s+/ /;  # collapse whitespace for compare
        next if $anum =~ /^#[0-9]/;  # macro definitions
        my $linenum = pos_linenum($_, $+[0]);
        unless ($anum =~ /^A[0-9]{6,7}$/) {
          bad($linenum, "bad A-number: $anum");
          next;
        }

        if (defined (my $elem = $anum_to_annotation{$anum})) {
          my ($prev_annotation, $prev_linenum) = @$elem;
          if ($annotation ne $prev_annotation) {
            bad($linenum, "$anum annotation \"$annotation\"");
            bad($prev_linenum, "  ... but previously \"$prev_annotation\"");
          }
        } else {
          $anum_to_annotation{$anum} = [$annotation, $linenum];
        }
        if ($annotation eq '') {
          push @unannotated, "$tex_filename:$linenum: unannotated $anum\n";
        }

        if ($annotation ne '') {
          if (defined (my $elem = $annotation_to_anum{$annotation})) {
            my ($prev_anum, $prev_linenum) = @$elem;
            if ($anum ne $prev_anum) {
              bad($linenum, "$anum annotation \"$annotation\"");
              bad($prev_linenum, "  ... duplicate of previous $prev_anum");
            }
          } else {
            $annotation_to_anum{$annotation} = [$anum, $linenum];
          }
        }
      }
    }
    my $distinct = scalar(keys %anum_to_annotation);
    if (@unannotated < 10) {
      print @unannotated;
    }
    my $unannotated = grep {$_->[0] eq ''} values %anum_to_annotation;
    print "$distinct distinct A-numbers, with $unannotated unannotated\n";
  }

  #----------------------------------------------------------------------------
  # check \index contents
  # - text@tex parts same
  # - if any $ or \ then it's tex and must have @

  verbose(2, "\\index contents");
  for ($tex_str_sans_comments) {
    my %index;
    my $braces = $RE{'balanced'}{-parens=>'{}'};
    while (m<\\index($braces)>g) {
      my $start = $-[0];
      my $whole = $1;
      $whole = substr($whole,1,-1);
      ### $whole

      if ($whole =~ /[\$\\]/) {
        my $texism = $&;
        if ($whole !~ /\@/) {
          my $linenum = pos_linenum($_, $start);
          bad($linenum, "\\index with $texism TeX but without \@ separator");
          print "  $whole\n";
        }
      }

      foreach my $part (split /!/, $whole) {
        my @ats = split /\@/, $part;
        if (@ats < 2) { next; }
        if (@ats > 2) {
          my $linenum = pos_linenum($_, $start);
          bad($linenum, "\\index multiple \@s: $whole");
          print "   part: $part\n";
          next;
        }
        my ($plain,$tex) = @ats;
        $tex =~ s/\\penalty\d+//g;         # line breaks
        $tex =~ s/\\['"]//g;               # accents
        $tex =~ s/\|[()]//;                # |( and |) index range
        $tex =~ s/\\hyp(\{\}| )?/-/g;      # hypenat \hyp{}
        $tex =~ s/\\MyTightDots/.../g;
        $tex =~ s/\\omega(\b|_)/w/g;       # sort \omega as "w"
        $tex =~ s/\\Omega(\b|_)\b/Omega/g; # sort \Omega as "Omega"
        $tex =~ s/\\alpha\b/a0/g;          # sort \alpha as "a0" so first
        $tex =~ s/\\phi\b/p0/g;            # sort \phi as "p0" so first
        $tex =~ s/\\leftrightarrow/<->/g;
        $tex =~ s{\\[td]?frac(\d)(\d)}{$1/$2}g;
        $tex =~ s/$kern_re//go;
        $tex =~ s/\\[a-zA-Z]+|[\${}_]//g;  # other macros and specials
        $plain =~ s/^000-//;         # hack for A-numbers
        $plain =~ s/^0+([1-9])/$1/;  # leading 0s for numeric ordering
        $plain =~ s/[ \n]//g;
        $tex =~ s/[ \n]//g;
        unless ($plain eq $tex) {
          my $linenum = pos_linenum($_, $start);
          bad($linenum, "\\index different plain/tex: $whole");
          print "   part: $part\n";
          print "   crunched plain: $plain\n";
          print "   crunched tex:   $tex\n";
        }
      }
    }
  }

  #----------------------------------------------------------------------------
  # check "{=} -" or "{=} +" which is the sign as operator
  #
  # but \phantom{=} ok?  (?<!\\phantom)

  verbose (2, "{=} followed by + or -");
  for ($tex_str_sans_comments) {
    while (/\{=\}\s*(?<operator>[-+]|\\(pm|mp))/g) {
      my $operator = $+{'operator'};
      my $linenum = pos_linenum($_);
      bad($linenum, "{=} with following operator: $operator");
    }
  }

  #----------------------------------------------------------------------------
  # LocalWords ispell

  verbose (2, "LocalWords usage");
  {
    tie my %LocalWords, 'Tie::IxHash';  # so report in document order
    for ($tex_str) {
      while (/LocalWords:(.*)/g) {
        my $words = $1;
        my $linenum = pos_linenum($_);
        $words =~ s/^[ \t]+|[ \t]+$//g;  # leading and trailing whitespace
        foreach my $word (split / +/, $words) {
          if (exists $LocalWords{$word}) {
            bad($linenum, "duplicate LocalWords: $word");
          } else {
            $LocalWords{$word} = $linenum;
          }
        }
      }
    }
    print scalar(keys %LocalWords)," LocalWords\n";
    my $str = $tex_str_sans_comments;
    $str =~ s<\\[-`'"~]><>g;        # sans accents and optional hyphen
    $str =~ s<\\[Hcv]\{(.)\}><$1>g;  # \c{X}
    $str =~ s<\\([lL])\{\}><$1>g;  # \l{} or \L{}
    foreach my $word (keys %LocalWords) {
      unless ($str =~ /[^[:alpha:]]\Q$word\E[^[:alpha:]]/) {
        bad_but_not_fatal($LocalWords{$word}, "unused LocalWords: $word");
      }
    }
    undef $str;
  }

  #----------------------------------------------------------------------------
  # \smash{} outside cases or aligned

  verbose (2, "\\smash");
  {
    my %allow_environments = (cases => 1,
                              aligned => 1,
                              tikzpicture => 1);
    for ($tex_str) {
      my @within;
    MATCH: while (/%.*|\\(begin|end)\{([^}]*)\}|\\phantom\{\\smash|\\smash\b(.*)/g) {
        my $begin_end = $1;
        my $environment = $2;
        my $post_smash = $3;
        my $startpos = $+[0];
        if ($begin_end) {
          if ($begin_end eq 'begin') {
            push @within, $environment;
          } else {
            if (@within && $within[-1] eq $environment) {
              pop @within;
            } else {
              my $linenum = pos_linenum($_, $startpos);
              bad($linenum, "unbalanced begin/end: $environment");
            }
          }
        } elsif (defined $post_smash) {
          unless ($post_smash =~ /ALLOW/) {
            foreach my $within (@within) {
              if ($allow_environments{$within}) {
                next MATCH;
              }
            }
            my $linenum = pos_linenum($_, $startpos);
            bad($linenum, "\\smash outside cases or aligned");
            # print $post_smash,"\n";
          }
        }
      }
    }
  }

  #----------------------------------------------------------------------------
  # empty \\ line break

  verbose(2, "empty consecutive \\\\ line breaks");
  for ($tex_str_sans_comments) {
    while (/\\\\[ \t\r\n]+\\\\/g) {
      my $linenum = pos_linenum($_, $+[0]);
      bad($linenum, "empty consecutive \\\\ line breaks");
    }
  }

  #----------------------------------------------------------------------------
  # \begin{theorem}\label{} should be label name "theorem-foo"
  # \MyFigureText\label{} should be label name "figure-foo"
  # Helps catch \autoref vs \eqref mistakes.

  for ($tex_str) {
    # {theorem}\label{foo}
    while (/\{theorem\}\\label\{(?<label>[^}]*)\}/g) {
      my $label = $+{'label'};
      unless ($label =~ /^theorem-/) {
        my $linenum = pos_linenum($_, $+[0]);
        bad($linenum, "theorem label should be \"theorem-...\"");
      }
    }
  }
  for ($tex_str) {
    # \MyFigureText\label{foo}
    # \caption{text}\label{foo}
    while (/(\\MyFigureText|\\caption\{[^}]*\}\s*)
            \\label\{(?<label>[^}]*)\}/sgx) {
      my $label = $+{'label'};
      unless ($label =~ /^figure-/) {
        my $linenum = pos_linenum($_, $+[0]);
        bad($linenum, "figure label name should be \"figure-...\"");
      }
    }
  }

  #----------------------------------------------------------------------------
  # OEIS Axxxxxx_samples OFFSET claimed

  verbose(2, "A-number OFFSET claimed");
  for ($tex_str) {
    while (/(A\d{6,7})_samples.*OFFSET\s*=\s*(-?\d+)/g) {
      my $anum = $1;
      my $offset = $2;
      ### match: "$anum OFFSET=$offset"
      my $linenum = pos_linenum($_, $+[0]);
      my $seq = eval { Math::NumSeq::OEIS::File->new (anum => $anum) };
      if (! defined $seq) {
        my $error = $@;
        $error =~ s/\n*$//;
        bad($linenum, $error);
      } elsif ($seq->i_start != $offset) {
        bad($linenum,
            "$anum shown OFFSET=$offset but sequence is actually ",
            $seq->i_start);
      }
    }
  }

  if (0) {
    require Devel::FindBlessedRefs;
    my %seen;
    Devel::FindBlessedRefs::find_refs_by_coderef
        (sub {
           my ($obj) = @_;
           my $class = Scalar::Util::blessed($obj) || return;
           if ($obj->isa('Math::NumSeq')) {
           }
             my $addr = Scalar::Util::refaddr ($obj);
             $seen{$class}->{$addr} = 1;
         });
    my @classes = sort keys %seen;
    my $is_toplevel_class = sub {
      my ($class, $class_list) = @_;
      return ! List::Util::first {$class ne $_ && $class->isa($_)} @$class_list;
    };
    my $traverse;
    $traverse = sub {
      my ($depth, $class_list) = @_;
      my @toplevels = grep {$is_toplevel_class->($_,$class_list)} @$class_list;
      foreach my $class (@toplevels) {
        my $count = scalar keys %{$seen{$class}};
        printf "%*s%s %d\n", 2*$depth, '', $class, $count;
        my @subclasses = grep {$_ ne $class && $_->isa($class)} @$class_list;
        $traverse->($depth+1, \@subclasses);
      }
    };
    $traverse->(1, \@classes);
  }

  #----------------------------------------------------------------------------
  # check for % GP-DEFINE redefinitions

  verbose(2, "GP-DEFINE redefinitions");
  {
    my $line_to_define = sub {
      my ($gp) = @_;
      ### line_to_define(): $gp
      if ($gp =~ /([A-Za-z0-9_]+)\s*=\s*memoize\(\1\s*\)/) {
        ### foo=memoize(foo) ...
        return undef;
      }
      if ($gp =~ m{^% GP-(CONSTANT|VECTOR|MATRIX)\s+(\w+)}) {
        return $2;
      }
      if ($gp =~ m{^\s*$gp_comment_re}o) {
        ### comment ...
        return undef;
      }
      while ($gp =~ /(?<builtin> for|forstep|my)\(
                   |(?<define>  [a-z0-9_]+)(\(.*\))?\s*=/ixg) {
        ### match ...
        ### builtin: $+{'builtin'}
        ### define : $+{'define'}
        last if defined $+{'builtin'};
        return $+{'define'};
      }
      return undef;
    };
    my %redefinition_ok = ();
    my $str = $tex_str;
    my $continuation = 0;
    my $in_brace = 0;
    my $define = '';   # name of the current function definition
    my $count_defines = 0;
    my %seen;
    while ($str =~ /^% (Test-Pari|GP)-(DEFINE\s+(?<gp>.*)|(CONSTANT|VECTOR|MATRIX)\s+(?<constant>\w+))/img) {
      my $this = $+{'constant'};
      my $gp = $+{'gp'};
      ### constant: $this
      if (defined $gp) {
        $gp =~ s/^\s+//;  # delete leading whitespace
        ### gp line: $gp
        ### $continuation
        ### $in_brace
        unless ($continuation || $in_brace) {
          $this = $line_to_define->($gp)
        }
      }
      if (defined $this) {
        ### define of: $this
        $count_defines++;
        my $linenum = pos_linenum($str);
        if (defined $seen{$this}) {
          bad_but_not_fatal($linenum, "GP-DEFINE redefinition: $this");
          bad_but_not_fatal($seen{$this}, "  ... previous here");
        } else {
          $seen{$this} = $linenum;
        }
        if (defined $gp) {
          $define = $this;
        }
      }

      if (defined $gp) {
        $continuation = ($gp =~ /\\\s*$/);
        if ($gp =~ /\{/) { $in_brace = 1; }
        if ($gp =~ /\}/) { $in_brace = 0; }

        # error("foo() is blah blah ...")
        if ($gp =~ /error\(\"(oops, )?(.*)\(\)/) {
          my $error_name = $2;
          ### error message from : $error_name
          if ($error_name ne $define) {
            my $linenum = pos_linenum($str);
            bad($linenum, "error() message about \"$error_name\" but in function \"$define\"");
          }
        }
      }
    }
    print "$count_defines gp-inline functions and variables\n";
  }

  #----------------------------------------------------------------------------
  # OEIS Axxxxxx_samples without an OFFSET recorded

  # Not sure always care about this ...
  # {
  #   my $str = $tex_str;
  #   while ($str =~ /\\\n% GP-Test\s+(A\d{6,7})_samples(.*OFFSET=)?/g) {
  #     my $anum = $1;
  #     my $offset = $2;
  #     unless (defined $offset) {
  #       my $linenum = pos_linenum($str, $+[0]);
  #       bad($linenum, "${anum}_samples without intended OFFSET recorded");
  #     }
  #   }
  # }

  #----------------------------------------------------------------------------
  # tikzpicture various
  # no bare node or coordinate after ;
  #
  # No need for ; after \foreach like \foreach \x in { ... }; <---
  # This shows up as "Missing character: There is no ; in font nullfont!"
  # in the .log file (like other unrecognised stuff in tikzpicture).
  #
  # Do need ; at the end of \path foreach ... {}; the same as other \path
  # of course.

  verbose(2, "tikzpicture");
  {
    # These tikz keys are mutually exclusive: any 2 appearing together is
    # bad (will use the last one appearing, though offsets accumulate, or
    # something like that).
    my %exclusive_keys = (left => 1,
                          right => 1,
                          above => 1,
                          below => 1,
                          'above right' => 1,
                          'above left' => 1,
                          'below right' => 1,
                          'below left' => 1,
                          anchor => 1);

    for ($tex_str_sans_comments) {
      while (/\\begin\{tikzpicture\}(.*?)\\end\{tikzpicture\}/sg) {
        my $str_linenum = pos_linenum($_, $-[1]) - 1;
        my $tikz = $1;

        # excluding ;; here, which is picked up above separately everywhere
        while ($tikz =~ /;\s*([^[:space:]\\}&;]\w*)|$braces_re/sgo) {
          if (defined (my $word = $1)) {
            my $linenum = $str_linenum + pos_linenum($tikz);
            bad($linenum, "tikzpicture ; followed by non-macro: $word");
          } else {
            # print "skip $&\n";
          }
        }
        
        while ($tikz =~ /\\foreach\s+\\\w+\s+in[^{]*?($braces_re\s*)+;/sgo) {
          my $linenum = $str_linenum + pos_linenum($tikz);
          bad($linenum, "tikzpicture no need for ; after \\foreach");
        }
        
        # no multiple right,left,anchor etc
        while ($tikz =~ /\[(.*?)\]/sg) {
          my $args = $1;
          my $linenum = $str_linenum + pos_linenum($tikz);
          my $seen;
          ### $args
          foreach my $key (split /,\s*/, $args) {
            # ENHANCE-ME: Allow arg value balanced parens like actual pgf parse.
            $key =~ s/=.*//s;
            ### $key
            next unless $exclusive_keys{$key};
            if ($seen) {
              bad($linenum, "tikzpicture both $seen and $key is unclear");
            } else {
              $seen = $key;
            }
          }
        }

        {
          # align=left etc without \\ using that
          my $tikz_trimmed = $tikz;
          $tikz_trimmed =~ s{\.style=$braces_re}{str_leave_newlines($&)}ges;
          while ($tikz_trimmed =~ /\balign=(.*?)(?<!\\);/sg) {
            my $rest = $1;
            next if defined $2;
            my $linenum = $str_linenum + pos_linenum($tikz_trimmed, $-[0]);
            unless ($rest =~ /\\\\|#\d/) {
              bad($linenum, "align= without \\\\ using it");
            }
          }
        }
      }
    }
  }


  #----------------------------------------------------------------------------
  # mathtools showonlyrefs=true and showonlyrefs=false should alternate,
  # not repeat

  verbose(2, "mathtools showonlyrefs alternating");
  for ($tex_str_sans_comments) {
    my $prev;
    my $prev_linenum;
    while (/showonlyrefs=(true|false)/g) {
      my $this = $1;
      my $linenum = pos_linenum($_, $+[0]);
      if (defined $prev && $this eq $prev) {
        bad($linenum, "showonlyrefs is already $this");
        bad($prev_linenum, "  ... previous setting here");
      } else {
        $prev = $this;
        $prev_linenum = $linenum;
      }
    }
  }


  #----------------------------------------------------------------------------
  # align etc no \tag and \notag together
  # showonlyrefs=false want always either \label or \notag

  {
    my $str = $tex_str_sans_comments;

    # strip \begin{cases} ... \end{cases} etc
    $str =~ s{\\begin\{(cases|aligned|pmatrix|matrix\*?|tikzpicture)\}.*?\\end\{\1\}}
             {str_leave_newlines($&)}ges;

    my $within = 'end';
    my %seen;
    my $showonlyrefs = 'false';
    my $showonlyrefs_location = 'default';
    while ($str =~ /(?<separator>\\\\
                    |\\intertext
                    |\\(?<within>begin|end)\{(align|equation|gather)\})
                  |(?<showonlyrefs>showonlyrefs(=(?<showonlyrefsvalue>true|false))?)
                  |\\(?<got>tag\*?|notag|label)
                   /xg) {
      if (defined $+{'separator'}) {
        ### separator: $+{'separator'}
        if ($within eq 'begin') {
          my $total = sum(0, values %seen);
          if ($showonlyrefs eq 'false' && $total != 1) {
            my $linenum = pos_linenum($str, $+[0]);
            bad($linenum, "in showonlyrefs=$showonlyrefs have $total but want exactly 1 of \\label, \\tag, \\notag");
            bad($showonlyrefs_location, "  showonlyrefs is here ...");
          } elsif ($total > 1) {
            my $linenum = pos_linenum($str, $+[0]);
            bad($linenum, "duplicates: ".join(',', sort keys %seen));
          }
        }
        if (defined $+{'within'}) {
          $within = $+{'within'};
        }
        %seen = ();
      } elsif (defined $+{'showonlyrefs'}) {
        $showonlyrefs = $+{'showonlyrefsvalue'} // 'true';
        $showonlyrefs_location = "line " . pos_linenum($str, $+[0]);
        ### $showonlyrefs
      } else {
        my $got = $+{'got'};
        $got =~ s/\*$//;
        $seen{$got}++;
        ### got: $+{'got'}
      }
    }
  }


  #----------------------------------------------------------------------------
  # prefer \\[] space measures to have blank line before (blank comment line)
  # including \\*[]

  {
    my $space_measure_non_negative_re = qr/\\\\\\*?\[[^-]/;

    # Matches a line which is not an empty comment "%".
    # So matches a non-comment line, or matches a comment line which is not
    # empty.
    my $not_an_empty_comment = qr/|[^%].*|%-*[^-\n].*/;

    for ($tex_str) {
      while (/\n$not_an_empty_comment\n$space_measure_non_negative_re/go){
        my $spacing = $2;
        my $linenum = pos_linenum($_, $+[0]);
        bad($linenum, "prefer blank comment line before \\\\[] spacing");
      }
    }
  }

  #----------------------------------------------------------------------------
  # check paired \FloatBarrier around \begin{figure} or \begin{my-plain-float}

  {
    my %want_next = (none               => 'FloatBarrier',
                     'FloatBarrier'     => 'begin',
                     begin              => 'end',
                     end                => 'FloatBarrier.end',
                     'FloatBarrier.end' => 'FloatBarrier',
                    );

    for my $str ($tex_str_sans_comments) {
      my $prev_env;
      my $prev = 'none';
      while ($str =~ m{\\((?<barrier>FloatBarrier)
                       |(?<env>begin|end)\{my-plain-float\}
                       )
                    }xg) {
        ### match: $&
        my $this;
        if (defined $+{'barrier'}) {
          $this = $+{'barrier'};
          if (defined $prev_env) {
            $this .= ".$prev_env";
          }
          undef $prev_env;
        } else {
          $this = $+{'env'};
          $prev_env = $this;
        }
        ### $prev
        ### $this
        unless ($want_next{$prev} eq $this) {
          my $linenum = pos_linenum($str);
          bad($linenum, "this $this prev $prev but wanted $want_next{$prev}");
        }
        $prev = $this;
      }
    }
  }


  #----------------------------------------------------------------------------
  # check for % generated by FILENAME exists

  verbose (2, "generator programs existence");
  for ($tex_str) {
    my $dirname = File::Basename::dirname($tex_filename);
    my %seen;
    while (/% generated by (\S+)/img) {
      my $filename = $1;
      $filename =~ s/,$//;
      $seen{$filename} = 1;
      my $filename_here = File::Spec->catfile($dirname,$filename);
      my $filename_up   = File::Spec->catfile($dirname, File::Spec->updir,
                                              $filename);
      if (! -e $filename_here && ! -e $filename_up) {
        my $linenum = pos_linenum($_);
        bad_but_not_fatal($linenum, "generated by $filename: does not exist");
      }
    }

    foreach my $basename (sort(File::Slurp::read_dir
                               (File::Spec->catdir ($dirname,'tools')))) {
      next unless $basename =~ /^gen-/;
      my $filename = "tools/$basename";
      if (! $seen{$filename}) {
        print "$filename:1: unused generator program\n";
        # $bad = 1;  # not fatal
      }
    }

    print scalar(keys %seen)," generator programs\n";
  }


  #----------------------------------------------------------------------------
  # \index{} in maths mode can come out in .idx with macros spaced out, for
  # example
  #
  #     \begin{align}
  #     \index{foo@\textit{foo}}
  #     \end{align}
  #     =>
  #     \indexentry{foo@\textit  {foo}}{1}     % in .idx file
  #
  # Check here that a \index{} of the same contents is consistently either
  # always in maths mode or always toplevel.  Maths is recognised as
  # "align", "gathered", and \mbox{\index{...}}.  The latter is a good way
  # to force a toplevel to be in maths, if you would otherwise have a
  # mixture.

  verbose (2, "\\index{} in maths");
  for ($tex_str_sans_comments) {
    my $contained_type = 'toplevel';
    my %begin_end_to_type = (begin => 'in-maths',
                             end   => 'toplevel');
    my %seen;
    while (m{\\(?<begin>begin|end)\{(align|gather)\}
             |(?<mbox>\\mbox\{)?\\index(?<index>$braces_re)
            }xsgo) {
      if (defined $+{'begin'}) {
        $contained_type = $begin_end_to_type{$+{'begin'}} or die "oops";
        next;
      }
      my $this_type = $contained_type;
      if ($+{'mbox'}) {
        $this_type = 'in-maths';
      }
      my $index = $+{'index'} // die "oops";
      $index =~ s/^\{//;
      $index =~ s/\}$//;
      $index =~ s/\|[()]$//;
      next unless $index =~ /\\/;

      my $linenum = pos_linenum($_);
      if (my $seen = $seen{$index}) {
        my ($seen_type,$seen_linenum) = @$seen;
        if ($seen_type ne $this_type) {
          bad($linenum, "\\index $this_type for: $index");
          bad($seen_linenum, "  but previous $seen_type will be different to makeindex");
        }
      } else {
        $seen{$index} = [$this_type,$linenum];
      }
    }
  }


  #----------------------------------------------------------------------------
  # check \mathtoolsset preceded by % comment or blank line,
  # so text doesn't flow to it

  for ($tex_str) {
    while (/[^\n]\\mathtoolsset/g) {
      my $prev = $1 // '';
      my $linenum = pos_linenum($_, $-[0]);
      bad($linenum, "mathtoolsset not at start of line");
    }
    while (/(.*)\n\\mathtools/g) {
      my $prev = $1;
      my $linenum = pos_linenum($_, $+[0]);
      next if $prev =~ /^\s*$/;
      next if $prev =~ /^%/;
      next if $prev =~ /^\\(end|usepackage|vspace)/;
      bad($linenum, "mathtoolsset prefer preceding comment line so no reflow by Emacs");
    }
  }

  #----------------------------------------------------------------------------
  # check ;; doubled semis, usually PGF or GP unnecessary extras

  verbose (2, "double semicolons ;;");
  for ($tex_str) {
    while (/;{2,}/g) {
      my $whole = $&;
      my $linenum = pos_linenum($_, $-[0]);
      bad($linenum, "multi-semicolons $whole");
    }
  }

  #----------------------------------------------------------------------------
  # any == >= etc
  # Exceptions made for PGF style
  #     shorten <=.1em       line ends
  #     shorten >=.1em
  #     >=Latex              arrow types
  #     >=Stealth
  #     >={Stealth}
  # But == can occur in tikzmaths so have to exclude all tikz probably.

  verbose (2, "possible C or GP operators");
  for ($tex_str_sans_comments_and_tikz) {
    while (/==
           |(?<!\{)(?<!shorten\ )[<>]=(?!Stealth)(?!\{)
           |=>/xg) {
      my $op = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "apparent C or GP etc operator: $op");
    }
  }

  #----------------------------------------------------------------------------
  # \end{document} duplicate

  verbose (2, "duplicate \\end{document}");
  for ($tex_str_sans_comments) {
    my $first_linenum;
    while (/\\end\{document\}/g) {
      my $linenum = pos_linenum($_);
      if (defined $first_linenum) {
        bad($linenum, "\\end{document} duplicate");
        bad($first_linenum, "  ... first is here");
      } else {
        $first_linenum = $linenum;
      }
    }
  }

  #----------------------------------------------------------------------------
  # Long GP-Test lines

  # Many currently ...
  # {
  #   my $str = $tex_str;
  #   while ($str =~ /^% GP-Test .{72}/mg) {
  #     my $linenum = pos_linenum($str, $+[0]);
  #     bad($linenum, "long GP-Test line");
  #   }
  # }


  #----------------------------------------------------------------------------
  # Double-quote " other than in comments, and other than \" accent.

  verbose (2, "double-quote \"");
  for ($tex_str_sans_comments) {
    while (/(\")|\\./g) {
      if ($1) {
        my $linenum = pos_linenum($_);
        bad($linenum, "double-quote \" should be `` and ''");
      }
    }
  }

  #----------------------------------------------------------------------------
  # \\begin{scope} duplicate keys
  # some shift= repeats accumulate, other settings override, maybe

  if (0) {
    for ($tex_str_sans_comments) {
      while (/\\begin\{scope\}\s*\[([^]]*)/osg) {
        my $params = $1;
        my $pos = $-[0];
        # print "$params\n";
        my %seen;
        while ($params =~ /\s*([^=,]+)(=([^({,]|\(.*?\)|\{.*?\})*)?/osg) {
          my $pos = $pos + $-[1];
          my $key = $1;
          my $value = $2 // '';
          $key =~ s/^,\s*//;
          # print "  key   $key\n";
          # print "  value $value\n";
          if ($seen{$key}++) {
            my $linenum = pos_linenum($_, $pos);
            bad($linenum, "tikz scope duplicate key: $key");
          }
        }
      }
    }
  }

  #----------------------------------------------------------------------------
  # \hskip probably prefer \hspace in LaTeX
  # safer against a "plus whatever" after
  # but \hspace causes a new box or word boundary somehow ...

  # verbose (2, "\\hskip prefer \\hspace");
  # for ($tex_str) {
  #   while (/\\hskip\b/g) {
  #     my $linenum = pos_linenum($_, $+[0]);
  #     bad($linenum, "prefer \\hspace over \\hskip, safer against "plus whatever"");
  #   }
  # }

  #----------------------------------------------------------------------------
  # \hspace{} after tikzpicture sometimes needs \hbox{\hspace{}}, it seems

  verbose (2, "\\hspace after tikzpicture");
  for ($tex_str) {
    my $space_measure_non_negative_re = qr/\\\\\[[^-]/;
    while (/\\end\{tikzpicture\}\n\\(hspace|kern)/g) {
      my $linenum = pos_linenum($_, $+[0]);
      bad($linenum, "may need \\hbox{\\hspace{}} for post-tikz bias");
    }
  }

  #----------------------------------------------------------------------------
  # no tab characters

  verbose (2, "no tab characters");
  for ($tex_str) {
    while (/\t(.*\t)*/g) {
      my $linenum = pos_linenum($_);
      bad($linenum, "tab character, prefer spaces over tabs");
    }
  }

  #----------------------------------------------------------------------------
  # check \mathit contents
  # eg. \mathit{foo(x)} should be \mathit{foo}(x)

  verbose (2, "\\mathit contents");
  for my $str ($tex_str_sans_comments) {
    my $linebreak_re = qr{\\linebreak\b(\[[0-4]\])?};
    my $mathit_allowed = qr{^([a-zA-Z0-9, ]
                            |\{,\}        # {,} unspaced comma
                            |\{\}         # {} against kerning
                            |\\[ ]        # \ space
                            |$kern_re\b
                            |$linebreak_re
                            |\\(alpha|delta)\b
                            |\\MyNegativePhantom\{\\/\\/\}
                            |\\MyTightDots[ ]
                            )*$}xo;
    my %index;
    while ($str =~ m/mathit($braces_re)/go) {
      my $mathit = $1;
      $mathit =~ s/^\{//;
      $mathit =~ s/\}$//;
      if ($mathit eq '') {
        my $linenum = pos_linenum($str);
        bad($linenum, "empty \\mathit");
      } elsif ($mathit =~ /mathit/) {
        my $linenum = pos_linenum($str);
        bad($linenum, "mathit within mathit: $mathit");
      } elsif ($mathit !~ $mathit_allowed) {
        my $linenum = pos_linenum($str);
        bad($linenum, "bad chars in \\mathit: $mathit");
      }
    }
  }


  #----------------------------------------------------------------------------
  # \end{align} followed by \index{}

  verbose (2, "{align} environment followed by \\index");
  for ($tex_str_sans_comments) {
    while (/\\end\{align\}.*\n?\\index/g) {

      my $linenum = pos_linenum($_); # at the \index part
      bad($linenum, "\\end{} followed by \\index{} is bad spacing (put it inside the align)");
    }
  }

  #----------------------------------------------------------------------------
  # check \label{foo-formula} outside "align"

  verbose(2, "\\label{foo-formula} outside {align}");
  for ($tex_str_sans_comments) {
    my $within_align = 0;
    my $count_labels = 0;
    my $align_envname_re = qr/(fl)?align/;
    while (m<\\(
               (?<Begin> begin)
             |(?<End>    end\{$align_envname_re})
             |(?<My>     MyEquationTag\\)?
               label\{(?<Label> [^}]*)
             )>xgo) {
      ### begin: $+{'Begin'}
      if (defined $+{'Begin'}) {
        ### begin align ...
        $within_align = 1;
      } elsif (defined $+{'End'}) {
        ### end align ...
        $within_align = 0;
      } elsif (defined $+{'My'}) {
        ### MyEquationTag ...
        # \MyEquationTag\label{foo} is ok
      } else {
        my $label = $+{'Label'};
        ### $label
        $count_labels++;
        if (! $within_align
            && $label =~ /formula/
            && $label !~ /^page-/) {
          my $linenum = pos_linenum($_);
          bad($linenum, "formula label outside maths: $label");
        }
      }
    }
    verbose (1, "$count_labels labels");
  }

  #----------------------------------------------------------------------------
  # check \begin{align} etc ending with a \\ separator

  verbose(2, "{align} environment ending \\\\");
  for ($tex_str_sans_comments_and_tikz) {
    my $whitespace_or_kern_re = qr/$comment_re
                                  |\s*|$kern_re
                                  |\\hspace$braces_re/ox;
    while (m/\\\\\
             (\[[^]]*])?
             ($whitespace_or_kern_re)*
             \\end\{(?<Type>align|equation|cases)
            /xmgo) {
      my $type = $+{'Type'};
      my $linenum = pos_linenum($_, $-[0]); # at start
      bad($linenum, "trailing \\\\ in $type");
    }
  }

  #----------------------------------------------------------------------------
  # any \pagebreak reckoned as a leftover

  for ($tex_str_sans_comments) {
    while (/\\pagebreak(?<option>\[\d+\])?(?<allow>\s+% ALLOW)?/g) {
      next if defined $-{'option'};   # \pagebreak[0] is ok
      next if defined $-{'allow'};    # "... % ALLOW" is ok
      my $linenum = pos_linenum($_);
      bad($linenum, "\\pagebreak left-over");
    }
  }


  #----------------------------------------------------------------------------
  # check for lower x. x want \space

  for ($tex_str_sans_comments_and_tikz) {
    while (/[a-z][)']*[.?!] [^ \t]/g) {
      my $word = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "no-end-sentence should be \\ (backslash space): $word");
    }
  }

  # check for abbreviations want .\space
  for ($tex_str_sans_comments_and_tikz) {
    while (/(ie|i\.e)\.(?!\\)/g) {
      my $word = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "abbreviation without \\space: $word");
    }
  }
  #----------------------------------------------------------------------------
  # superscript or subscript of paren not brace

  verbose(2, "superscript or subscript paren");
  for ($tex_str_sans_comments) {
    while (/[_^][][()]/g) {
      my $whole = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "subscript/superscript should be braces");
    }
  }

  #----------------------------------------------------------------------------
  # \sqrt99 etc multiple alnum chars after \sqrt should be \sqrt{..}

  for ($tex_str_sans_comments) {
    while (/\\sqrt[0-9][[:alnum:]]+/g) {
      my $whole = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "\\sqrt of multiple chars should have braces");
    }
  }

  #----------------------------------------------------------------------------
  # question mark ? with whitespace before,
  # sometimes from cut and paste of gp output

  verbose(2, "question mark");
  for ($tex_str_sans_comments) {
    while (/[ \t\r\n]\?/g) {
      my $linenum = pos_linenum($_);
      bad($linenum, "question mark ? with blank before");
    }
  }

  #----------------------------------------------------------------------------
  # OEIS should have \textsc{} when in plain text

  # NO
  # for ($tex_str_sans_comments) {
  #   while (/(\\textsc\{|title=|-)?\bOEIS\b/g) {
  #     next if defined $1;
  #     my $linenum = pos_linenum($_, $+[0]);
  #     bad($linenum, "$& without \\textsc");
  #   }
  # }

  #----------------------------------------------------------------------------
  # preferring smallcaps OEIS

  # NO
  # for ($tex_str_sans_comments) {
  #   while (/\\textsc\{[^}]*}|\b(OEIS)\b/g) {
  #     my $word = $2 // next;
  #     my $linenum = pos_linenum($_, $-[0]);
  #     bad($linenum, "prefer \\textsc{$word}");
  #   }
  # }

  #----------------------------------------------------------------------------
  # check unbalanced lfloor / rfloor
  #
  # occasionally have unbalanced lparen/rparen when padding

  verbose(2, "balanced lfloor / rfloor");
  {
    my $str = $tex_str;
    str_strip_phantom($str);
    my @stack;
    while ($str =~ m{(?<comment> $comments_re)
                   |
                     \\
                     ( (?<dir>    [rl])
                       (?<type>   ceil|floor|brace)
                     | (?<dotdir> left|right)\.
                     )
                     (?<unclosed> \s*%\s*(unclosed|no\ opening))?
                  }xgo) {
      if (defined $+{'comment'}) {
        if ($+{'comment'} =~ /ALLOW-UNCLOSED/) {
          pop @stack;
        }
        next;
      }
      my $whole = $&;
      my $dir = $+{'dir'};
      my $type = $+{'type'};
      my $dotdir = $+{'dotdir'};
      my $unclosed = $+{'unclosed'};
      my $linenum = pos_linenum($str);
      ### $whole
      ### $unclosed
      ### $linenum

      next if defined $unclosed;
      if (defined $dotdir) {
        if ($dotdir eq 'left') {
          push @stack, ['.', $linenum];
        } else {
          if (! @stack) {
            bad($linenum, "$whole without matching left");
            next;
          }
          pop @stack;  # \right. matches anything
        }
      } elsif ($dir eq 'l') {
        push @stack, [$type, $linenum];
      } elsif ($dir eq 'r') {
        if (! @stack) {
          bad($linenum, "$whole without matching left");
          next;
        }
        my $elem = pop @stack;
        my ($opp_type,$opp_linenum) = @$elem;
        unless ($opp_type eq $type || $opp_type eq '.') {
          bad($opp_linenum, "unbalanced types $whole");
          bad($linenum,     "  from l$opp_type here");
        }
      } else {
        die;
      }
    }
    while (@stack) {
      my $elem = pop @stack;
      my ($type,$linenum) = @$elem;
      bad($linenum, "unclosed l$type");
    }
  }

  #----------------------------------------------------------------------------
  # check no commented or disguised GP-Test etc

  verbose(2, "disguised GP-Test");
  {
    my $comment_prefix_re
      = qr{^[ \t]*
           (?:
             [\#%*]+    # # Perl, % TeX, * C continuing
           |//+         # // C++
           |/\*+        # /* C
           |=for[ \t]   # =for Perl POD
           )?           # or nothing
        }mx;
    for ($tex_str) {
      while (/$comment_prefix_re[ \t]*((.*?)GP-[-A-Za-z0-9]+)/mgo) {
        my $whole = $1;
        my $disguise = $2;
        next unless $disguise;
        next if $disguise eq 'The ';
        my $linenum = pos_linenum($_, $-[0]);
        ### $&
        ### $whole
        ### $disguise
        ### $linenum
        bad($linenum, "disguised gp-inline: $whole");
      }
    }
  }

  #----------------------------------------------------------------------------
  # \textit{S} of long word should have \hyphenation

  verbose(2, "\\textit hyphenation");
  for ($tex_str_sans_comments) {
    my %builtin_hyphenation = (direction => 1);
    my %seen;
    while (/\\textit\{(.*?)\}/g) {
      my $textit = $1;
      my $linenum = pos_linenum($_, $-[0]);
      foreach my $word (split /\W+/, $textit) {
        next if $builtin_hyphenation{$word};
        next if $hyphenation{$word};
        next if length($word) < 5;
        next if $seen{$word}++;
        bad($linenum, "textit without \\hyphenation: $word");
      }
    }
  }

  #----------------------------------------------------------------------------
  # long words should have \hyphenation

  if (0)        # too many
    {
      my $str = $tex_str_sans_comments_and_tikz;
      $str =~ s<^.*?\\begin\{document\}><str_leave_newlines($&)>es;
      $str =~ s/\\(mathit|url|hyphenation|label|eqref|noeqref|cite|nocite|bibitem)\{.*?\}//g;
      $str =~ s/\\index\{.*?@(.*?)\}/$1/g;
      $str =~ s<\\(begin|end)\{.*?\}><>g;

      my %builtin_hyphenation = (direction => 1);
      require TeX::Hyphen;
      my $hyp = TeX::Hyphen->new
        (file => '/usr/share/texlive/texmf-dist/tex/generic/hyphen/hyphen.tex');
      my %seen;
      while ($str =~ /(\\)?([[:alpha:]]+)/g) {
        next if $1;
        my $word = $2;
        next if length($word) < 8;
        next if $seen{$word}++;
        my $linenum = pos_linenum($str, $-[0]);
        my @splits = $hyp->hyphenate($word);
        next if @splits >= 2;
        bad_but_not_fatal($linenum, "long word without \\hyphenation: $word");

        $word = lc($word);
        if ($word =~ s/(s|ing|ings)$//) {
          my $end = $1;
          my $visualize = $hyp->visualize($word) . $end;
          if ($visualize =~ /-/) {
            print "    maybe: \\hyphenation{$visualize}\n";
          }
        }
      }
    }

  #----------------------------------------------------------------------------
  # check unused hyphenations

  verbose(2, "unused hyphenations");
  for ($tex_str_sans_comments) {
    my %unused = %hyphenation;
    while (/\w+/g) {
      my $word = $&;
      if (defined $unused{$word}) {
        delete $unused{$word};
        last unless %unused;
      }
    }
    foreach my $word (sort {$hyphenation{$a} <=> $hyphenation{$b}}
                      keys %unused) {
      my $linenum = $hyphenation{$word};
      bad_but_not_fatal($linenum, "unused \\hyphenation: $word");
    }
  }

  #----------------------------------------------------------------------------
  # prefer \operatorname{Re} and Im rather than \mathit{}

  verbose(2, "\\operatorname{Re}");
  for ($tex_str_sans_comments) {
    while (/\\mathit\{(Re|Im)\}/ig) {
      my $whole = $&;
      my $inside = $1;
      my $linenum = pos_linenum($_, $-[0]);
      bad($linenum, "$inside prefer \\operatorname: $whole");
    }
  }

  #----------------------------------------------------------------------------
  # check \mathit{S} of a single character

  verbose(2, "\\mathit single char");
  for ($tex_str_sans_comments) {
    while (/\\mathit\{.\}/g) {
      my $mathit = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "mathit of single character: $mathit");
    }
  }

  #----------------------------------------------------------------------------
  # check \dotsc not preceded by , comma

  verbose(2, "\\dotsc preceded by comma");
  {
    my $str = $tex_str_sans_comments;
    $str =~ s/\\phantom\{\\dotsc\}//g;
    $str =~ s/\\linebreak(\[\d+\])?//g;
    $str =~ s/\\math[rlc]lap\{(\\dotsc)\}/$1/g;
    while ($str =~ m/\\dotsc/g) {
      # print "dotsc at ",pos($str),"\n";
      my $beg = $-[0];
      my $len = min($beg,256);
      my $preceding = substr($str,$beg-$len,$len);
      my $whole = '';
      while ($preceding =~ m<(\s+
                             |\\(degree
                               |[,;]
                               |hspace\{[^}]*\}
                               |$kern_sans_backslash_re
                               )
                             )$>xo) {
        $whole = $& . $whole;
        $preceding = substr($preceding,0,$-[0]);
      }
      unless ($preceding =~ /(,|[0-9]*\.[0-9]+)$/) {
        my $linenum = pos_linenum($str);
        bad($linenum, "\\dotsc not preceded by comma");    # : $whole
      }
    }
  }

  #----------------------------------------------------------------------------
  # check unused \label{}

  for my $str ($tex_str_sans_comments) {

    tie my %labels, 'Tie::IxHash';  # in the order seen in the document
    while ($str =~ m<\\label($braces_re)>g) {
      my $label = $1;
      next if $label =~ /#/;  # skip labels generated with macro parameters
      my $linenum = pos_linenum($str);
      if (defined $labels{$label}) {
        bad($linenum, "duplicate label: $label");
        bad($labels{$label}, "was defined here");
      } else {
        $labels{$label} = $linenum;
      }
    }

    my %used_labels;
    while ($str =~ m<\b(ref|eqref|noeqref|autoref|autopageref)\*?$braces_re>g) {
      my $label = $2;
      next if $label =~ /#/;  # macro parameters
      if (! exists $labels{$label}) {
        my $linenum = pos_linenum($str);
        bad($linenum, "no such label: $label");
      }
      $used_labels{$label} = 1;
    }

    # # labels used in comments are allowed
    # while ($tex_str =~ m<(eqref|noeqref|autoref|autopageref)\*?$braces_re>g) {
    #   my $label = $2;
    #   next if $label =~ /#/;  # macro parameters
    #   $used_labels{$label} = 1;
    # }

    foreach my $label (keys %labels) {
      if (! $used_labels{$label}) {
        my $linenum = $labels{$label};
        bad_but_not_fatal($linenum, "unused label: $label");
      }
    }
  }

  #----------------------------------------------------------------------------
  # check for % GP-DEFINE braces { } with backslashing too

  for ($tex_str) {
    my %redefinition_ok = ('for' => 1,
                           'my'  => 1);
    my $in_brace = 0;
    while (/^% GP-DEFINE(.*)/img) {
      my $gp = $1;
      if ($gp =~ /\{/) { $in_brace = 1; }
      if ($in_brace && $gp =~ /\\\s*$/) {
        my $linenum = pos_linenum($_);
        bad($linenum, "GP-DEFINE backslash in braces: $gp");
      }
      if ($gp =~ /\}/) { $in_brace = 0; }
    }
  }

  #----------------------------------------------------------------------------
  # \pagebreak[0] should be at the end of a paragraph

  for ($tex_str) {
    while (m/\\pagebreak
             (\[[^]]*])?       # optional [0] etc
             [ \t]*
             ${comments_re}
             [ \t]*
             (?<after>\S*)
            /xsg) {
      my $after = $+{'after'};
      next if $after =~ /^\\begin\{/ || $after eq '';
      my $linenum = pos_linenum($_, $-[0]);
      bad($linenum, "\\pagebreak not at end of paragraph");
      print "  after: $after\n";
    }
  }

  #----------------------------------------------------------------------------
  # check [my dot] (nodename)

  # {
  #   my $str = $tex_str_sans_comments;
  #
  #   while ($str =~ m/ dot\b[^]\{]*]\s*\(/g) {
  #     my $type = $+{'type'};
  #     my $linenum = pos_linenum($str, $-[0]);
  #     bad($linenum, "[my dot] (nodename)");
  #   }
  # }

  #----------------------------------------------------------------------------
  # check unbackslashed kern

  verbose(2, "unbackslashed kern");
  for ($tex_str) {
    while (/(?<![\\m])kern[-.0-9]|(?<![\\])mkern[-.0-9]/g) {
      my $kern = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "unbackslashed kern: $kern");
    }
  }

  #----------------------------------------------------------------------------
  # check doubled words

  verbose(2, "doubled words");
  {
    # >= 1 whitespace chars, at most 1 newline
    my $whitespace_max_one_newline_re
      = qr/[ \t\r]+(?:\n[ \t\r]*)?|[ \t\r]*\n[ \t\r]+/;

    my $str = $tex_str_sans_comments;

    # delete \begin{tabular}{c l l l}  and {longtable}
    $str =~ s/\{(tabular|longtable)\}.*//g;

    # word of 2 or more chars, separated by whitespace
    while ($str =~ /\b([a-z]\w*)$whitespace_max_one_newline_re(\1)\b/sgio) {
      my $doublon = $1;
      my $other = $2;
      # case insensitive on first char only, rest exact to be bad
      next unless substr($doublon,1) eq substr($other,1);

      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "doubled word: $doublon $other");
    }
    undef $str
  }

  #----------------------------------------------------------------------------
  # check for % GP-Test ending in ;

  verbose(2, "GP-Test ending semicolon ;");
  for ($tex_str) {
    while (/^% (Test-Pari|GP-Test)\s+.*;\s*$/img) {
      my $gp_test = $1;
      my $linenum = pos_linenum($_);
      bad($linenum, "$gp_test end with ;");
    }
  }

  #----------------------------------------------------------------------------
  # check for indented % GP-Test etc

  for ($tex_str) {
    while (/^[ \t]+% ((Test-Pari|GP-)\S*)/mg) {
      my $gp = $1;
      my $linenum = pos_linenum($_);
      bad($linenum, "indented $gp");
    }
  }

  #----------------------------------------------------------------------------
  # check for \frac123 more than 2 digits

  for ($tex_str) {
    while (/\\[td]?frac\d{3,}/g) {
      my $frac = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "$frac more than 2 digits");
    }
  }

  #----------------------------------------------------------------------------
  # check for \eqref{figure- or theorem- as they should be \autoref
  # check for \autoref{...-formula}      as it should be \eqref

  for ($tex_str) {
    while (/\\autoref\{[^}]*?-formula\}/g) {
      my $all = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "$all should be \\eqref");
    }
  }
  for ($tex_str) {
    while (/\\eqref\{(figure|theorem)-.*?\}/g) {
      my $all = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "$all should be \\autoref");
    }
  }

  #----------------------------------------------------------------------------
  # check for hyphens which should use \hyp{}

  verbose(2, "hypens with \\hyp{}");
  {
    my $str = $tex_str_sans_comments_and_tikz;

    # no label part of \eqref{foo-label-xxx}
    # keep [] part of \cite[text text]{foo-label-xxx}
    $str =~ s/(
                \\(hyphenation
                |bibitem|label|tagform|newtagform
                |(cite|ref|eqref|noeqref|autoref|autopageref)\*?
                |url|href|hypertarget|hyperlink
                |(sub)*section
                |MyDOI
                )
                (\[.*?\])?         # optional [foo] extra
              )
              \{[^}]*\}            # label name, not subject to \hyp{}
             /$1/gx;
    $str =~ s/\\usetagform\{.*?\}//g;   # \usetagform{} name
    $str =~ s/\\newfloat\{.*?\}//g;     # \newfloat{} environ name
    $str =~ s/\\(begin|end)\{.*?\}//g;  # \begin{foo} environ names
    $str =~ s/(\\index\{)[^@}]*@/$1/g;  # \index{text@tex} text part skip
    $str =~ s/(\\(index|makeindex|printindex))\[.*?\]/$1/sg;  # index filenames
    $str = str_strip_MyWebPage($str);
    $str =~ s/pdfkeywords=\{[^}]*\}//g;
    $str =~ s/\\MyFurtherProof\{[^}]*}\{[^}]*}//g;

    while ($str =~ /(\\)?([[:alnum:]]+)-([[:alnum:]]+)/ig) {
      next if defined $1;  # macro with - after is ok
      my $pre = $2;
      my $post = $3;
      my $word = $&;
      my $linenum = pos_linenum($str);
      pos($str) = $-[3];   # to consider $post part and further dash

      # 6 or more digits might by hyphenated
      if ($pre =~ /^[[:digit:]]*$/ && $post =~ /^[[:digit:]]*$/) {
        next unless length($pre) >= 6 || length($post) >= 6;
      }

      # 5 or more letters might by hyphenated
      next unless length($pre) >= 5 || length($post) >= 5;

      bad($linenum, "hyphen should use \\hyp{}: ",$word);
    }
  }

  for ($tex_str_sans_comments) {
    while (/\\hyp\{\}-/g) {
      my $linenum = pos_linenum($_);
      bad($linenum, "\\hyp{} with dash");
    }
  }

  #----------------------------------------------------------------------------
  # check \newcommand{\Foo}{123}; ending semicolon

  for ($tex_str_sans_comments) {
    while (/.*(\\newcommand.*\};)[ \t]*(%.*)?$/mg) {
      my $command = $&;
      my $linenum = pos_linenum($_);
      my $col = $+[1] - $-[0] + 1;
      bad("$linenum:$col", "\\newcommand ending semicolon: $command");
    }
  }

  #----------------------------------------------------------------------------
  # check for capital X. missing X\@.
  #
  # ? ! : get extra space like end-of-sentence characters.
  # Think : is ok left alone.
  # Right paren or quote ) ' skipped.
  # TeX Book page 74 ...

  for ($tex_str_sans_comments_and_tikz) {
    while (/([[:alnum:]]*[A-Z][)']*)[.?!]([ \t][ \t]|\n)/g) {
      my $word = $1;
      my $linenum = pos_linenum($_,$-[2]);
      bad($linenum, "end-of-sentence should be: $word\\\@.");
    }
  }

  #----------------------------------------------------------------------------
  # check for sentence ending ".."

  {
    my $str = $tex_str_sans_comments_and_tikz;
    $str =~ s/(\\index\{)[^}]*?(@)/$1$2/g;  # kill plain text part of \index{}
    while ($str =~ /[.,;]\./g) {
      my $dots = $&;
      my $linenum = pos_linenum($str);
      bad($linenum, "two dots: $dots");
    }
  }

  #----------------------------------------------------------------------------
  # check for ",," anywhere except comments (which can be gp comma stuff)

  for ($tex_str_sans_comments) {
    while (/,\s*,/sg) {
      my $commas = $&;
      my $linenum = pos_linenum($_);
      bad($linenum, "two commas: $commas");
    }
  }

  #----------------------------------------------------------------------------
  # check for xx/yy with / instead of \slash or similar
  # two or more chars on each side
  # Some false positives on tikz /sqrt(2) and similar.

  {
    my $str = $tex_str_sans_comments_and_tikz;
    str_strip_MyWebPage($str);
    $str =~ s/(\\(url|href)\{)[^}]*(\})/$1$3/g; # delete guts of \url or \href
    $str =~ s/(\(\$).*?(\$\))/$1 $2/g;  # delete guts of ($ .. $)

    while ($str =~ m{[a-z][a-z]+/([a-z][a-z]+)}ig) {
      my $word = $&;
      my $linenum = pos_linenum($str);
      pos($str) = $-[1];
      bad($linenum, "/ rather than \\slash or similar: $word");
    }
  }

  verbose(2, "end");
}

#------------------------------------------------------------------------------
sub pos_linenum {
  my ($str, $pos) = @_;
  $pos //= pos($_[0]);
  $str = substr($str, 0, $pos);
  return 1 + scalar($str =~ tr/\n//);
}

sub count_lines {
  my ($str) = @_;
  my $count = ($str =~ tr/\n/\n/);
  if ($str =~ /[^\n]$/) { $count++; }
  return $count;
}

# return $str with all characters except \n newlines deleted
sub str_leave_newlines {
  my ($str) = @_;
  $str =~ tr/\n//cd;
  return $str;
}

sub str_strip_MyWebPage {
  $_[0] =~ s/\\newcommand\{?\\MyWebPage[^\n]*//;
}

# FIXME: backslashed braces are not recognised, but much prefer \lbrace for
# clarity anyway
sub str_strip_phantom {
  $_[0]  =~ s<\\v?phantom$braces_re>
             <str_leave_newlines($&)>eg;
}

#------------------------------------------------------------------------------

{
  my $help = sub {
    print "$progname [--options] filename...\n";
    my @opts =
      (['-h, --help',    'Print this help'],
       ['-v, --version', 'Print program version'],
       ['--verbose',     'Print extra messages'],
      );
    my $width = 2 + max (map { length ($_->[0]) } @opts);
    foreach (@opts) {
      printf "%-*s%s\n", $width, $_->[0], $_->[1];
    }
    print "\n";
    exit 0;
  };

  GetOptions ('help|?' => $help,
              version => sub {
                print "$progname version $VERSION\n";
                exit 0;
              },
              verbose    => \$verbose,
             )
    or exit 1;

  @ARGV or $help->();
}

while (@ARGV) {
  $tex_filename = shift @ARGV;
  check_file();
}

exit ($bad ? 1 : 0);
